import logging
from .. import entities

logger = logging.getLogger(name='dtlpy')


class BaseSample:
    def __init__(self,
                 start_time,
                 end_time,
                 project_id,
                 org_id,
                 pipeline_id,
                 event_type,
                 action,
                 status,
                 other_keys: dict = None
                 ):
        self.start_time = start_time
        self.end_time = end_time
        self.project_id = project_id
        self.org_id = org_id
        self.pipeline_id = pipeline_id
        self.event_type = event_type
        self.action = action
        self.status = status
        self.other_keys = other_keys

    def to_json(self):
        _json = {
            'startTime': self.start_time,
            'endTime': self.end_time,
            'context': {
                'projectId': self.project_id,
                'projectOrgId': self.org_id,
                'pipelineId': self.pipeline_id
            },
            'eventType': self.event_type,
            'action': self.action,
            'data': {
                'status': self.status
            }
        }
        if self.other_keys is not None:
            if 'context' in self.other_keys:
                _json['context'].update(self.other_keys['context'])
            if 'data' in self.other_keys:
                _json['data'].update(self.other_keys['data'])
            _json.update({k: v for k, v in self.other_keys.items() if k not in ['context', 'data']})

        return _json


class ServiceSample(BaseSample):
    def __init__(self,
                 start_time=None,
                 end_time=None,
                 user_id=None,
                 project_id=None,
                 org_id=None,
                 pipeline_id=None,
                 pipeline_node_id=None,
                 service_id=None,
                 pod_id=None,
                 pod_type=None,
                 event_type=None,
                 entity_type=None,
                 action=None,
                 status=None,
                 num_restarts=None,
                 cpu=None,
                 ram=None,
                 queue_size=None,
                 num_executions=None,
                 service_type: entities.ServiceType = None,
                 interval: int = None,
                 driver_id: str = None,
                 other_keys: dict = None,
                 concurrency_limit: int = None,
                 concurrency_count: int = None,
                 cpu_limit: int = None,
                 ram_limit: int = None,
                 gpu_limit: int = None,
                 gpu: int = None,
                 gpu_memory: int = None,
                 gpu_memory_limit: int = None
                 ):
        super().__init__(
            start_time=start_time,
            end_time=end_time,
            project_id=project_id,
            org_id=org_id,
            pipeline_id=pipeline_id,
            event_type=event_type,
            action=action,
            status=status,
            other_keys=other_keys
        )
        self.user_id = user_id
        self.pipeline_node_id = pipeline_node_id
        self.service_id = service_id
        self.pod_id = pod_id
        self.pod_type = pod_type
        self.entity_type = entity_type
        self.num_restarts = num_restarts
        self.cpu = cpu
        self.ram = ram
        self.queue_size = queue_size
        self.num_executions = num_executions
        self.service_type = service_type if service_type is not None else entities.ServiceType.REGULAR
        self.interval = interval
        self.driver_id = driver_id
        self.concurrency_limit = concurrency_limit
        self.concurrency_count = concurrency_count
        self.cpu_limit = cpu_limit
        self.ram_limit = ram_limit
        self.gpu_limit = gpu_limit
        self.gpu = gpu
        self.gpu_memory = gpu_memory
        self.gpu_memory_limit = gpu_memory_limit

    def to_json(self):
        _json = super().to_json()
        _json['context'].update({
            'userId': self.user_id,
            'pipelineNodeId': self.pipeline_node_id,
            'serviceId': self.service_id,
            'podId': self.pod_id,
            'podType': self.pod_type,
            'serviceType': self.service_type
        })
        _json['data'].update({
            'numRestarts': self.num_restarts,
            'cpu': self.cpu,
            'ram': self.ram,
            'queueSize': self.queue_size,
            'numExecutions': self.num_executions,
            'interval': self.interval,
            'driverId': self.driver_id,
            'concurrencyLimit': self.concurrency_limit,
            'concurrencyCount': self.concurrency_count,
            'cpuLimit': self.cpu_limit,
            'ramLimit': self.ram_limit,
            'gpuLimit': self.gpu_limit,
            'gpu': self.gpu,
            'gpuMemory': self.gpu_memory,
            'gpuMemoryLimit': self.gpu_memory_limit
        })
        _json.update({
            'entityType': self.entity_type
        })
        _json['context'] = {k: v for k, v in _json['context'].items() if v is not None}
        _json['data'] = {k: v for k, v in _json['data'].items() if v is not None}
        return {k: v for k, v in _json.items() if v is not None}

    @classmethod
    def from_json(cls, _json):
        inst = cls(
            start_time=_json.get('startTime', None),
            end_time=_json.get('endTime', None),
            user_id=_json.get('context', {}).get('userId', None),
            project_id=_json.get('context', {}).get('projectId', None),
            org_id=_json.get('context', {}).get('projectOrgId', None),
            pipeline_id=_json.get('context', {}).get('pipelineId', None),
            pipeline_node_id=_json.get('context', {}).get('pipelineNodeId', None),
            service_id=_json.get('context', {}).get('serviceId', None),
            pod_id=_json.get('context', {}).get('podId', None),
            pod_type=_json.get('context', {}).get('podType', None),
            event_type=_json.get('eventType', None),
            entity_type=_json.get('EntityType', None),
            action=_json.get('action', None),
            status=_json.get('data', {}).get('status', None),
            num_restarts=_json.get('data', {}).get('numRestarts', None),
            cpu=_json.get('data', {}).get('cpu', None),
            ram=_json.get('data', {}).get('ram', None),
            queue_size=_json.get('data', {}).get('queueSize', None),
            num_executions=_json.get('data', {}).get('numExecutions', None),
            service_type=_json.get('type', entities.ServiceType.REGULAR),
            interval=_json.get('data', {}).get('interval', None),
            driver_id=_json.get('data', {}).get('driverId', None),
            concurrency_limit=_json.get('data', {}).get('concurrencyLimit', None),
            concurrency_count=_json.get('data', {}).get('concurrencyCount', None),
            cpu_limit=_json.get('data', {}).get('cpuLimit', None),
            ram_limit=_json.get('data', {}).get('ramLimit', None),
            gpu_limit=_json.get('data', {}).get('gpuLimit', None),
            gpu=_json.get('data', {}).get('gpu', None),
            gpu_memory=_json.get('data', {}).get('gpuMemory', None),
            gpu_memory_limit=_json.get('data', {}).get('gpuMemoryLimit', None)
        )
        return inst


class ExecutionSample(BaseSample):
    def __init__(self,
                 start_time=None,
                 end_time=None,
                 project_id=None,
                 org_id=None,
                 pipeline_id=None,
                 event_type=None,
                 action=None,
                 status=None,
                 user_id=None,
                 account_id=None,
                 pipeline_node_id=None,
                 pipeline_execution_id=None,
                 service_id=None,
                 execution_id=None,
                 trigger_id=None,
                 function_name=None,
                 duration=None,
                 other_keys: dict = None,
                 function_duration=None
                 ):
        super().__init__(
            start_time=start_time,
            end_time=end_time,
            project_id=project_id,
            org_id=org_id,
            pipeline_id=pipeline_id,
            event_type=event_type,
            action=action,
            status=status,
            other_keys=other_keys
        )
        self.user_id = user_id
        self.account_id = account_id
        self.pipeline_node_id = pipeline_node_id
        self.pipeline_execution_id = pipeline_execution_id
        self.service_id = service_id
        self.execution_id = execution_id
        self.trigger_id = trigger_id
        self.function_name = function_name
        self.duration = duration
        self.function_duration = function_duration

    def to_json(self):
        _json = super().to_json()
        _json['context'].update({
            'userId': self.user_id,
            'accountId': self.account_id,
            'pipelineNodeId': self.pipeline_node_id,
            'pipelineExecutionId': self.pipeline_execution_id,
            'serviceId': self.service_id,
            'triggerId': self.trigger_id
        })
        _json['data'].update({
            'functionName': self.function_name,
            'duration': self.duration,
            'functionDuration': self.function_duration,
        })
        _json['context'] = {k: v for k, v in _json['context'].items() if v is not None}
        _json['data'] = {k: v for k, v in _json['data'].items() if v is not None}
        return {k: v for k, v in _json.items() if v is not None}

    @classmethod
    def from_json(cls, _json):
        inst = cls(
            start_time=_json.get('startTime', None),
            end_time=_json.get('endTime', None),
            user_id=_json.get('context', {}).get('userId', None),
            project_id=_json.get('context', {}).get('projectId', None),
            org_id=_json.get('context', {}).get('projectOrgId', None),
            account_id=_json.get('context', {}).get('accountId', None),
            pipeline_id=_json.get('context', {}).get('pipelineId', None),
            pipeline_node_id=_json.get('context', {}).get('pipelineNodeId', None),
            pipeline_execution_id=_json.get('context', {}).get('pipelineExecutionId', None),
            service_id=_json.get('context', {}).get('serviceId', None),
            execution_id=_json.get('context', {}).get('pipelineExecutionId', None),
            trigger_id=_json.get('context', {}).get('triggerId', None),
            event_type=_json.get('eventType', None),
            action=_json.get('action', None),
            status=_json.get('data', {}).get('status', None),
            function_name=_json.get('data', {}).get('functionName', None),
            duration=_json.get('data', {}).get('duration', None),
            function_duration=_json.get('data', {}).get('functionDuration', None)
        )
        return inst


class PipelineExecutionSample(BaseSample):
    def __init__(self,
                 start_time=None,
                 end_time=None,
                 project_id=None,
                 org_id=None,
                 pipeline_id=None,
                 event_type=None,
                 action=None,
                 status=None,
                 account_id=None,
                 pipeline_node_id=None,
                 pipeline_execution_id=None,
                 trigger_id=None,
                 node_status=None,
                 other_keys: dict = None
                 ):
        super().__init__(
            start_time=start_time,
            end_time=end_time,
            project_id=project_id,
            org_id=org_id,
            pipeline_id=pipeline_id,
            event_type=event_type,
            action=action,
            status=status,
            other_keys=other_keys
        )
        self.account_id = account_id
        self.pipeline_node_id = pipeline_node_id
        self.pipeline_execution_id = pipeline_execution_id
        self.trigger_id = trigger_id
        self.node_status = node_status

    def to_json(self):
        _json = super().to_json()
        _json['context'].update({
            'accountId': self.account_id,
            'pipelineNodeId': self.pipeline_node_id,
            'pipelineExecutionId': self.pipeline_execution_id,
            'triggerId': self.trigger_id
        })
        _json['data'].update({
            'nodeStatus': self.node_status
        })
        _json['context'] = {k: v for k, v in _json['context'].items() if v is not None}
        _json['data'] = {k: v for k, v in _json['data'].items() if v is not None}
        return {k: v for k, v in _json.items() if v is not None}

    @classmethod
    def from_json(cls, _json):
        inst = cls(
            start_time=_json.get('startTime', None),
            end_time=_json.get('endTime', None),
            project_id=_json.get('context', {}).get('projectId', None),
            org_id=_json.get('context', {}).get('projectOrgId', None),
            account_id=_json.get('context', {}).get('accountId', None),
            pipeline_id=_json.get('context', {}).get('pipelineId', None),
            pipeline_execution_id=_json.get('context', {}).get('pipelineExecutionId', None),
            trigger_id=_json.get('context', {}).get('triggerId', None),
            event_type=_json.get('eventType', None),
            action=_json.get('action', None),
            status=_json.get('data', {}).get('status', None),
            pipeline_node_id=_json.get('data', {}).get('pipelineNodeId', None),
            node_status=_json.get('data', {}).get('nodeStatus', None)
        )
        return inst



import logging
from .. import entities, repositories, exceptions, miscellaneous, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name="dtlpy")


def generate_basic_pipeline():
    return dict(name="", projectId="", nodes=[], connections=[], startNodes=[], variables=[])


class Pipelines:
    """
    Pipelines Repository

    The Pipelines class allows users to manage pipelines and their properties. See our documentation for more information on `pipelines <https://dataloop.ai/docs/pipelines-overview>`_.
    """

    def __init__(self, client_api: ApiClient, project: entities.Project = None):
        self._client_api = client_api
        self._project = project

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            try:
                self._project = repositories.Projects(client_api=self._client_api).get()
            except exceptions.NotFound:
                raise exceptions.PlatformException(
                    error="2001",
                    message='Missing "project". need to set a Project entity or use project.pipelines repository',
                )
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError("Must input a valid Project entity")
        self._project = project

    ###########
    # methods #
    ###########
    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/pipelines".format(self.project.id))

    def open_in_web(self, pipeline: entities.Pipeline = None, pipeline_id: str = None, pipeline_name: str = None):
        """
        Open the pipeline in web platform.

        **prerequisites**: Must be *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity
        :param str pipeline_id: pipeline id
        :param str pipeline_name: pipeline name

        **Example**:

        .. code-block:: python

            project.pipelines.open_in_web(pipeline_id='pipeline_id')
        """
        if pipeline_name is not None:
            pipeline = self.get(pipeline_name=pipeline_name)
        if pipeline is not None:
            pipeline.open_in_web()
        elif pipeline_id is not None:
            self._client_api._open_in_web(url=self.platform_url + "/" + str(pipeline_id))
        else:
            self._client_api._open_in_web(url=self.platform_url)

    @_api_reference.add(path="/pipelines/{pipelineId}", method="get")
    def get(self, pipeline_name=None, pipeline_id=None, fetch=None) -> entities.Pipeline:
        """
        Get Pipeline object to use in your code.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        You must provide at least ONE of the following params: pipeline_name, pipeline_id.

        :param str pipeline_id: pipeline id
        :param str pipeline_name: pipeline name
        :param fetch: optional - fetch entity from platform, default taken from cookie
        :return: Pipeline object
        :rtype: dtlpy.entities.pipeline.Pipeline

        **Example**:

        .. code-block:: python

            pipeline = project.pipelines.get(pipeline_id='pipeline_id')
        """
        if fetch is None:
            fetch = self._client_api.fetch_entities

        if pipeline_name is None and pipeline_id is None:
            raise exceptions.PlatformException(error="400", message="Must provide an identifier in inputs")
        elif fetch:
            if pipeline_id is not None:
                success, response = self._client_api.gen_request(
                    req_type="get", path="/pipelines/{}".format(pipeline_id)
                )
                if not success:
                    raise exceptions.PlatformException(response)
                pipeline = entities.Pipeline.from_json(
                    client_api=self._client_api, _json=response.json(), project=self._project
                )
                if pipeline_name is not None and pipeline.name != pipeline_name:
                    logger.warning(
                        "Mismatch found in pipeline.get: pipeline_name is different then pipeline.name:"
                        " {!r} != {!r}".format(pipeline_name, pipeline.name)
                    )
            elif pipeline_name is not None:
                filters = entities.Filters(
                    field="name", values=pipeline_name, resource=entities.FiltersResource.PIPELINE, use_defaults=False
                )
                if self._project is not None:
                    filters.add(field="projectId", values=self._project.id)
                pipelines = self.list(filters=filters)
                if pipelines.items_count == 0:
                    raise exceptions.PlatformException(
                        error="404", message="Pipeline not found. Name: {}".format(pipeline_name)
                    )
                elif pipelines.items_count > 1:
                    raise exceptions.PlatformException(
                        error="400",
                        message="More than one pipelines found by the name of: {} "
                        "Please get pipeline from a project entity".format(pipeline_name),
                    )
                pipeline = pipelines.items[0]
            else:
                raise exceptions.PlatformException(
                    error="400",
                    message="No checked-out pipeline was found, must checkout or provide an identifier in inputs",
                )
        else:
            pipeline = entities.Pipeline.from_json(
                _json={"id": pipeline_id, "name": pipeline_name},
                client_api=self._client_api,
                project=self._project,
                is_fetched=False,
            )

        return pipeline

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Pipeline]:
        pool = self._client_api.thread_pools(pool_name="entity.create")
        jobs = [None for _ in range(len(response_items))]

        for i_pipeline, pipeline in enumerate(response_items):
            jobs[i_pipeline] = pool.submit(
                entities.Pipeline._protected_from_json,
                **{"client_api": self._client_api, "_json": pipeline, "project": self._project}
            )

        # get all results
        # noinspection PyUnresolvedReferences
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        pipelines = miscellaneous.List([r[1] for r in results if r[0] is True])
        return pipelines

    def _list(self, filters: entities.Filters):
        url = "/pipelines/query"

        # request
        success, response = self._client_api.gen_request(req_type="post", path=url, json_req=filters.prepare())
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    @_api_reference.add(path="/pipelines/query", method="post")
    def list(self, filters: entities.Filters = None, project_id: str = None) -> entities.PagedEntities:
        """
        List project pipelines.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :param str project_id: project id
        :return: Paged entity
        :rtype: dtlpy.entities.paged_entities.PagedEntities

        **Example**:

        .. code-block:: python

            pipelines = project.pipelines.list()
        """
        if filters is None:
            filters = entities.Filters(resource=entities.FiltersResource.PIPELINE)
        # assert type filters
        elif not isinstance(filters, entities.Filters):
            raise exceptions.PlatformException(error="400", message="Unknown filters type: {!r}".format(type(filters)))
        if filters.resource != entities.FiltersResource.PIPELINE:
            raise exceptions.PlatformException(
                error="400",
                message="Filters resource must to be FiltersResource.PIPELINE. Got: {!r}".format(filters.resource),
            )

        if project_id is None and self._project is not None:
            project_id = self._project.id

        if project_id is not None:
            filters.add(field="projectId", values=project_id)

        paged = entities.PagedEntities(
            items_repository=self,
            filters=filters,
            page_offset=filters.page,
            page_size=filters.page_size,
            project_id=project_id,
            client_api=self._client_api,
        )
        paged.get_page()
        return paged

    def _name_validation(self, name: str):
        url = "/piper-misc/naming/packages/{}".format(name)

        # request
        success, response = self._client_api.gen_request(req_type="get", path=url)
        if not success:
            raise exceptions.PlatformException(response)

    # @_api_reference.add(path='/pipelines/{pipelineId}', method='delete')
    def delete(self, pipeline: entities.Pipeline = None, pipeline_name: str = None, pipeline_id: str = None):
        """
         Delete Pipeline object.

         **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity
        :param str pipeline_id: pipeline id
        :param str pipeline_name: pipeline name
        :return: True if success
        :rtype: bool

        **Example**:

         .. code-block:: python

             is_deleted = project.pipelines.delete(pipeline_id='pipeline_id')
        """
        # get id and name
        if pipeline_id is None:
            if pipeline is None:
                pipeline = self.get(pipeline_name=pipeline_name)
            pipeline_id = pipeline.id

        # request
        success, response = self._client_api.gen_request(req_type="delete", path="/pipelines/{}".format(pipeline_id))

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return results
        return True

    @_api_reference.add(path="/pipelines/{pipelineId}/settings", method="patch")
    def update_settings(self, pipeline: entities.Pipeline, settings: entities.PipelineSettings):
        """
        Update pipeline settings.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity
        :param dtlpy.entities.pipeline.PipelineSettings settings: settings entity
        :return: Pipeline object
        :rtype: dtlpy.entities.pipeline.Pipeline

        **Example**:

        .. code-block:: python

            pipeline = project.pipelines.update_settings(pipeline='pipeline_entity', settings=dl.PipelineSettings(keep_triggers_active=True))
        """
        # payload
        payload = {"settings": settings.to_json()}

        # request
        success, response = self._client_api.gen_request(
            req_type="patch", path="/pipelines/{}".format(pipeline.id), json_req=payload
        )
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Pipeline.from_json(_json=response.json(), client_api=self._client_api, project=self._project)

    def __update_variables(self, pipeline: entities.Pipeline):
        pipeline_json = pipeline.to_json()
        variables = pipeline_json.get("variables", list())

        for var in variables:
            if var.get("reference", None) is None:
                var["reference"] = pipeline.id

        # payload
        payload = {"variables": variables}

        # request
        success, response = self._client_api.gen_request(
            req_type="patch", path="/pipelines/{}/variables".format(pipeline.id), json_req=payload
        )
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Pipeline.from_json(_json=response.json(), client_api=self._client_api, project=self._project)

    @_api_reference.add(path="/pipelines/{pipelineId}", method="patch")
    def update(self, pipeline: entities.Pipeline = None) -> entities.Pipeline:
        """
        Update pipeline changes to platform.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity
        :return: Pipeline object
        :rtype: dtlpy.entities.pipeline.Pipeline

        **Example**:

        .. code-block:: python

            pipeline = project.pipelines.update(pipeline='pipeline_entity')
        """
        # payload
        payload = pipeline.to_json()

        # update settings
        if pipeline.settings_changed():
            new_pipeline = self.update_settings(pipeline=pipeline, settings=pipeline.settings)
            payload["settings"] = new_pipeline.to_json().get("settings", payload.get("settings"))

        # update variables
        if pipeline.variables_changed():
            new_pipeline = self.__update_variables(pipeline=pipeline)
            payload["variables"] = new_pipeline.to_json().get("variables", payload.get("variables"))

        success, response = self._client_api.gen_request(
            req_type="patch", path="/pipelines/{}".format(pipeline.id), json_req=payload
        )

        # exception handling
        if not success:
            raise exceptions.PlatformException(response)

        # return entity
        return entities.Pipeline.from_json(_json=response.json(), client_api=self._client_api, project=self._project)

    @_api_reference.add(path="/pipelines", method="post")
    def create(self, name: str = None, project_id: str = None, pipeline_json: dict = None) -> entities.Pipeline:
        """
        Create a new pipeline.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param str name: pipeline name
        :param str project_id: project id
        :param dict pipeline_json: json containing the pipeline fields
        :return: Pipeline object
        :rtype: dtlpy.entities.pipeline.Pipeline

        **Example**:

        .. code-block:: python

            pipeline = project.pipelines.create(name='pipeline_name')
        """
        if pipeline_json is None:
            pipeline_json = generate_basic_pipeline()

        if name is not None:
            pipeline_json["name"] = name

        if project_id is not None:
            pipeline_json["projectId"] = project_id
        else:
            if not pipeline_json.get("projectId", None):
                pipeline_json["projectId"] = self.project.id

        success, response = self._client_api.gen_request(req_type="post", path="/pipelines", json_req=pipeline_json)
        if success:
            pipeline = entities.Pipeline.from_json(
                client_api=self._client_api, _json=response.json(), project=self.project
            )
        else:
            raise exceptions.PlatformException(response)
        assert isinstance(pipeline, entities.Pipeline)
        return pipeline

    @_api_reference.add(path="/pipelines/{pipelineId}/install", method="post")
    def install(self, pipeline: entities.Pipeline = None, resume_option: entities.PipelineResumeOption = None):
        """
        Install (start) a pipeline.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity
        :param dtlpy.entities.pipeline.PipelineResumeOption resume_option: optional - resume pipeline method (what to do with existing cycles)
        :return: Composition object

        **Example**:

        .. code-block:: python

            project.pipelines.install(pipeline='pipeline_entity')
        """

        payload = {}
        if resume_option:
            payload["resumeOption"] = resume_option

        success, response = self._client_api.gen_request(
            req_type="post", path="/pipelines/{}/install".format(pipeline.id), json_req=payload
        )

        if not success:
            raise exceptions.PlatformException(response)

        return entities.Pipeline.from_json(client_api=self._client_api, _json=response.json(), project=self.project)

    @_api_reference.add(path="/pipelines/{pipelineId}/uninstall", method="post")
    def pause(self, pipeline: entities.Pipeline = None, keep_triggers_active: bool = None):
        """
        Pause a pipeline.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity
        :param bool keep_triggers_active: Do we want the triggers to stay active and collect events
        :return: Composition object

        **Example**:

        .. code-block:: python

            project.pipelines.pause(pipeline='pipeline_entity')
        """

        payload = {}
        if keep_triggers_active is not None:
            payload["keepTriggersActive"] = keep_triggers_active

        success, response = self._client_api.gen_request(
            req_type="post", path="/pipelines/{}/uninstall".format(pipeline.id), json_req=payload
        )

        if not success:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path="/pipelines/{pipelineId}/reset", method="post")
    def reset(
        self,
        pipeline: entities.Pipeline = None,
        pipeline_id: str = None,
        pipeline_name: str = None,
        stop_if_running: bool = False,
    ):
        """
        Reset pipeline counters.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity - optional
        :param str pipeline_id: pipeline_id -  optional
        :param str pipeline_name: pipeline_name -  optional
        :param bool stop_if_running: If the pipeline is installed it will stop the pipeline and reset the counters.
        :return: bool

        **Example**:

        .. code-block:: python

            success = project.pipelines.reset(pipeline='pipeline_entity')
        """

        if pipeline_id is None:
            if pipeline is None:
                if pipeline_name is not None:
                    pipeline = self.get(pipeline_name=pipeline_name)
                else:
                    raise exceptions.PlatformException(
                        "400", "Must provide one of pipeline, pipeline_id or pipeline_name"
                    )
            pipeline_id = pipeline.id

        if stop_if_running is True:
            if pipeline is None:
                pipeline = self.get(pipeline_id=pipeline_id)
            pipeline.pause()

        success, response = self._client_api.gen_request(
            req_type="post", path="/pipelines/{}/reset".format(pipeline_id)
        )

        if not success:
            raise exceptions.PlatformException(response)

        return True

    @_api_reference.add(path="/pipelines/{id}/statistics", method="get")
    def stats(self, pipeline: entities.Pipeline = None, pipeline_id: str = None, pipeline_name: str = None):
        """
        Get pipeline counters.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity - optional
        :param str pipeline_id: pipeline_id -  optional
        :param str pipeline_name: pipeline_name -  optional
        :return: PipelineStats
        :rtype: dtlpy.entities.pipeline.PipelineStats

        **Example**:

        .. code-block:: python

            pipeline_stats = project.pipelines.stats(pipeline='pipeline_entity')
        """

        if pipeline_id is None:
            if pipeline is None:
                if pipeline_name is not None:
                    pipeline = self.get(pipeline_name=pipeline_name)
                else:
                    raise exceptions.PlatformException(
                        "400", "Must provide one of pipeline, pipeline_id or pipeline_name"
                    )
            pipeline_id = pipeline.id

        success, response = self._client_api.gen_request(
            req_type="get", path="/pipelines/{}/statistics".format(pipeline_id)
        )

        if not success:
            raise exceptions.PlatformException(response)

        return entities.PipelineStats.from_json(_json=response.json())

    def execute(
        self,
        pipeline: entities.Pipeline = None,
        pipeline_id: str = None,
        pipeline_name: str = None,
        execution_input=None,
    ):
        """
        Execute a pipeline and return the pipeline execution as an object.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity
        :param str pipeline_id: pipeline id
        :param str pipeline_name: pipeline name
        :param execution_input: list of the dl.FunctionIO or dict of pipeline input - example {'item': 'item_id'}
        :return: entities.PipelineExecution object
        :rtype: dtlpy.entities.pipeline_execution.PipelineExecution

        **Example**:

        .. code-block:: python

            pipeline_execution= project.pipelines.execute(pipeline='pipeline_entity', execution_input= {'item': 'item_id'} )
        """

        if pipeline is None and pipeline_id is None:
            raise ValueError("Pipeline or pipeline_id must be provided")

        if pipeline is None:
            pipeline = self.get(pipeline_id=pipeline_id)

        execution = repositories.PipelineExecutions(
            pipeline=pipeline, client_api=self._client_api, project=self._project
        ).create(pipeline_id=pipeline.id, execution_input=execution_input, test_mode=False)
        return execution

    def test(self, pipeline: entities.Pipeline = None, pipeline_id: str = None, execution_input=None):
        """
        Execute a pipeline in test mode and return the pipeline execution as an object.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dtlpy.entities.pipeline.Pipeline pipeline: pipeline entity
        :param str pipeline_id: pipeline id
        :param execution_input: list of the dl.FunctionIO or dict of pipeline input - example {'item': 'item_id'}
        :return: entities.PipelineExecution object
        :rtype: dtlpy.entities.pipeline_execution.PipelineExecution

        **Example**:

        .. code-block:: python

            pipeline_execution= project.pipelines.test(pipeline='pipeline_entity', execution_input= {'item': 'item_id'} )
        """
        if pipeline is None and pipeline_id is None:
            raise ValueError("Pipeline or pipeline_id must be provided")

        if pipeline is None:
            pipeline = self.get(pipeline_id=pipeline_id)

        execution = repositories.PipelineExecutions(
            pipeline=pipeline, client_api=self._client_api, project=self._project
        ).create(pipeline_id=pipeline.id, execution_input=execution_input, test_mode=True)
        return execution

    @_api_reference.add(path="/pipelines/validate", method="post")
    def validate(self, pipeline_json: dict):
        """
        Validate a pipeline configuration.

        **prerequisites**: You must be an *owner* or *developer* to use this method.

        :param dict pipeline_json: pipeline configuration as JSON dictionary
        :return: Validation result
        :rtype: dict

        **Example**:

        .. code-block:: python

            validation_result = project.pipelines.validate(pipeline_json={'name': 'my_pipeline', 'nodes': []})
        """
        success, response = self._client_api.gen_request(
            req_type="post", path="/pipelines/validate", json_req=pipeline_json
        )

        if not success:
            raise exceptions.PlatformException(response)

        return response.json()



import datetime
import threading

from .. import entities


class BaseServiceRunner:
    _do_reset = False
    _auto_refresh_dtlpy_token = None
    _refresh_dtlpy_token = None
    _threads_terminated = list()
    _threads_terminated_lock = threading.Lock()
    _service_entity = None

    def do_reset(self):
        self._do_reset = True

    @staticmethod
    def ping(progress):
        progress.logger.debug('received ping at: {}'.format(datetime.datetime.now().isoformat()))
        return 'pong'

    def _terminate(self, tid):
        with self._threads_terminated_lock:
            self._threads_terminated.append(tid)

    def kill_event(self):
        ident = threading.get_ident()
        if ident in self._threads_terminated:
            with self._threads_terminated_lock:
                self._threads_terminated.pop(self._threads_terminated.index(ident))
            raise InterruptedError('Execution received termination signal')

    @property
    def service_entity(self) -> entities.Service:
        assert isinstance(self._service_entity, entities.Service), "service_entity must be a dl.Service object"
        return self._service_entity

    @service_entity.setter
    def service_entity(self, value):
        assert isinstance(value, entities.Service), "service_entity must be a dl.Service object"
        self._service_entity = value


class Progress:
    """
    Follow the event progress
    """

    def update(self,
               status=None,
               progress=0,
               message=None,
               output=None,
               duration=None,
               action=None
               ):
        """
        Update the progress flow

        :param str status: the progress status to display
        :param int progress: number of finished flow
        :param str message: the progress message to display
        :param dict output: json serializable object to update the event output
        :param float duration: the event duration
        :param str action: event action
        """
        pass


class ItemStatusEvent:
    def __init__(self, _json: dict):
        if _json is None:
            _json = dict()

        self.pipeline_id = _json.get('pipelineId', None)
        self.node_id = _json.get('nodeId', None)
        self.action = _json.get('action', None)

        status = _json.get('status', dict())
        if status is None:
            status = dict()

        self.task_id = status.get('taskId', None)
        self.assignment_id = status.get('assignmentId', None)
        self.status = status.get('status', None)
        self.creator = status.get('creator', None)
        self.timestamp = status.get('timestamp', None)


class ExecutionEventContext:
    def __init__(self, _json: dict):
        if _json is None:
            _json = dict()

        self.resource = _json.get('resource', None)
        self.source = _json.get('source', None)
        self.action = _json.get('action', None)
        self.resource_id = _json.get('resourceId', None)
        self.user_id = _json.get('userId', None)
        self.dataset_id = _json.get('datasetId', None)
        self.project_id = _json.get('projectId', None)
        self.body = _json.get('body', None)
        self.item_status_event = ItemStatusEvent(_json.get('itemStatusEvent', dict()))


    @property
    def command_data(self):
        if self.body is not None:
            return self.body.get('commandData', None)
        return None

    @property
    def command_id(self):
        if self.command_data is not None:
            return self.command_data.get('commandId', None)
        return None


class Context:
    """
    Contex of the service state
    """

    def __init__(
            self,
            service: entities.Service = None,
            package: entities.Package = None,
            project: entities.Project = None,
            event_context: dict = None,
            execution_dict: dict = None,
            progress: Progress = None,
            logger=None,
            sdk=None
    ):
        """
        A Context entity use DTLPY entities for context in a service runner


        :param dict execution_dict: the current execution dict in the state
        :param dict service: the current service entity in th state
        :param dict package: the current package entity in th state
        :param dict project: the current project entity in th state
        :param dict event_context: ExecutionEventContext json display the Execution event context
        :param dl.Progress progress: Progress object for work flow
        :param logger: logger object
        :param sdk: the dtlpy package
        """
        # dtlpy
        self._logger = logger
        self._sdk = sdk
        self._progress = progress

        self.event = ExecutionEventContext(event_context)
        if execution_dict is None:
            execution_dict = dict()
        self.execution_dict = execution_dict

        # ids
        self.trigger_id = execution_dict.get('trigger_id', None)
        self.execution_id = execution_dict.get('id', None)

        # pipeline
        pipeline = execution_dict.get('pipeline', dict())
        if pipeline is None:
            pipeline = dict()
        self.pipeline_id = pipeline.get('id', None)
        self.node_id = pipeline.get('nodeId', None)
        self.pipeline_execution_id = pipeline.get('executionId', None)

        # objects
        self._service = service
        self._package = package
        self._project = project
        self._task = None
        self._assignment = None
        self._pipeline = None
        self._node = None
        self._execution = None
        self._pipeline_execution = None

    @property
    def command_id(self):
        return self.event.command_id

    @property
    def command_data(self):
        return self.event.command_data

    @property
    def package(self):
        assert isinstance(self._package, entities.Package), "Missing `package` in context"
        return self._package

    @property
    def project(self):
        assert isinstance(self._project, entities.Project), "Missing `project` in context"
        return self._project

    @property
    def service(self):
        assert isinstance(self._service, entities.Service), "Missing `service` in context"
        return self._service

    @property
    def item_status_creator(self):
        return self.event.item_status_event.creator

    @property
    def item_status(self):
        return self.event.item_status_event.status

    @property
    def item_status_operation(self):
        return self.event.item_status_event.action

    @property
    def execution(self) -> entities.Execution:
        if self._execution is None:
            # noinspection PyProtectedMember
            self._execution = self.sdk.Execution.from_json(
                _json=self.execution_dict,
                client_api=self.service._client_api,
                service=self.service,
                project=self.project
            )
        return self._execution

    @property
    def task_id(self) -> str:
        return self.event.item_status_event.task_id

    @property
    def task(self) -> entities.Task:
        if self._task is None and self.task_id is not None:
            try:
                self._task = self.sdk.tasks.get(task_id=self.task_id)
            except Exception:
                self.logger.exception('Failed to get task')
        return self._task

    @property
    def assignment_id(self) -> str:
        return self.event.item_status_event.assignment_id

    @property
    def assignment(self) -> entities.Assignment:
        if self._assignment is None and self.assignment_id is not None:
            self._assignment = self.sdk.assignments.get(assignment_id=self.assignment_id)
        return self._assignment

    @property
    def pipeline(self) -> entities.Pipeline:
        if self._pipeline is None and self.pipeline_id is not None:
            self._pipeline = self.sdk.pipelines.get(pipeline_id=self.pipeline_id)
        return self._pipeline

    @property
    def node(self):
        if self._node is None and self.pipeline is not None and self.node_id is not None:
            self._node = [n for n in self.pipeline.nodes if n.node_id == self.node_id][0]
        return self._node

    @property
    def pipeline_execution(self):
        if self._pipeline_execution is None and self.pipeline_execution_id is not None:
            self._pipeline_execution = self.sdk.pipeline_executions.get(
                pipeline_execution_id=self.pipeline_execution_id,
                pipeline_id=self.pipeline_id
            )
        return self._pipeline_execution

    @property
    def sdk(self):
        if self._sdk is None:
            import dtlpy
            self._sdk = dtlpy
        return self._sdk

    @property
    def logger(self):
        if self._logger is None:
            import logging
            self._logger = logging.getLogger("[AgentContext]")
        return self._logger



def main():
    """
    Move items to another folder in platform
    :return:
    """
    import dtlpy as dl

    # Get project and dataset
    project = dl.projects.get(project_name='Ninja Turtles')
    dataset = project.datasets.get(dataset_name='Splinter')

    # Get all items from the source folder
    filters = dl.Filters()
    filters.add(field='filename', values='/fighting/**')  # take files from the directory only (recursive)
    filters.add(field='type', values='file')  # only files
    pages = dataset.items.list(filters=filters)

    dst_folder = '/fighting_shredder'
    # iterate through items
    for page in pages:
        for item in page:
            # move item
            item.move(new_path=dst_folder)



from . import BaseAnnotationDefinition


class Text(BaseAnnotationDefinition):
    """
        can create a text annotation having two types (paragraph, block)
    """
    type = "text_mark"

    def __init__(self, text_type, start, end, label,
                 top=None, left=None, attributes=None, description=None):
        """
        can create a text annotation having two types (paragraph, block)

        :param text_type: text type (paragraph, block)
        :param start: start position in characters
        :param end: end position in characters
        :param label: annotation label
        :param top: top box pixel
        :param left: left box pixel
        :param attributes: annotation attributes
        :param description:

        :return:
        """
        super().__init__(description=description, attributes=attributes)
        self.text_type = text_type
        self.start = start
        self.end = end
        self.top = top
        self.left = left
        self.label = label

    @property
    def x(self):
        return 0

    @property
    def y(self):
        return 0

    @property
    def geo(self):
        return list()

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        return self.add_text_to_image(image=image, annotation=self)

    def to_coordinates(self, color):
        coordinates = {
            "type": self.text_type,
            "label": self.label,
            "start": self.start,
            "end": self.end
        }
        if self.top is not None and self.left is not None:
            coordinates['top'] = self.top
            coordinates['left'] = self.left
        return coordinates

    @classmethod
    def from_json(cls, _json):
        coordinates = _json["coordinates"]

        return cls(
            text_type=coordinates.get("type"),
            start=coordinates.get("start"),
            end=coordinates.get("end"),
            top=coordinates.get("top", None),
            left=coordinates.get("left", None),
            label=_json["label"],
            attributes=_json.get("attributes", None),
        )



import dictdiffer
import logging
from typing import Union

logger = logging.getLogger(name='dtlpy')
TYPE = 0
FIELD = 1
LIST = 2


class DictDiffer:

    @staticmethod
    def diff(origin, modified):
        diffs = dict()

        dict_diff = list(dictdiffer.diff(origin, modified))
        for i_diff, diff in enumerate(dict_diff):
            modified_field_pointer = modified
            root_change = False
            if len(diff[FIELD]) > 0:
                field_pointer, modified_field_pointer, is_list = DictDiffer.get_field_path(
                    diffs=diffs,
                    path=diff[FIELD],
                    diff_type=diff[TYPE],
                    values=diff[LIST],
                    modified_field_pointer=modified_field_pointer
                )
            else:
                field_pointer, modified_field_pointer, is_list, root_change = origin, modified, False, True

            if is_list:
                for i in modified_field_pointer:
                    if i not in field_pointer:
                        field_pointer.append(i)
                continue
            else:
                if diff[TYPE] == 'add':
                    for addition in diff[LIST]:
                        field_pointer[addition[0]] = addition[1]
                        if root_change is True:
                            diffs[addition[0]] = addition[1]

                elif diff[TYPE] == 'remove':
                    for deletion in diff[LIST]:
                        field_pointer[deletion[0]] = None

                elif diff[TYPE] == 'change':
                    change = diff[LIST]
                    field = diff[FIELD]
                    if not isinstance(field, list):
                        field = field.split('.')
                    field_pointer[field[-1]] = change[1]
        return diffs

    @staticmethod
    def next_is_list(path: list, i_level: Union[str, int], values):
        if len(path) > (i_level + 1):
            if isinstance(path[i_level + 1], int):
                return True
        elif isinstance(values[0][0], int):
            return True
        return False

    @staticmethod
    def get_field_path(diffs, path, diff_type, modified_field_pointer, values):
        field_pointer = diffs
        if not isinstance(path, list):
            path = path.split('.')

        next_is_list = False
        if len(path) > 1 or diff_type != 'change':
            for i_level, level in enumerate(path):
                next_is_list = DictDiffer.next_is_list(i_level=i_level, path=path, values=values)
                if diff_type == 'change' and level == path[-2]:
                    if level not in field_pointer:
                        if next_is_list:
                            field_pointer[level] = list()
                        else:
                            field_pointer[level] = dict()
                    field_pointer = field_pointer[level]
                    modified_field_pointer = modified_field_pointer[level]
                    break
                if level not in field_pointer:
                    if next_is_list:
                        field_pointer[level] = list()
                    else:
                        field_pointer[level] = dict()
                field_pointer = field_pointer[level]
                modified_field_pointer = modified_field_pointer[level]

                if next_is_list:
                    break

        return field_pointer, modified_field_pointer, next_is_list



import dataclasses
import threading
import tempfile
import datetime
import logging
import string
import shutil
import random
import base64
import copy
import time
import tqdm
import traceback
import sys
import io
import os
from itertools import chain
from PIL import Image
from functools import partial
import numpy as np
from concurrent.futures import ThreadPoolExecutor
import attr
from collections.abc import MutableMapping
from typing import Optional
from .. import entities, utilities, repositories, exceptions
from ..services import service_defaults
from ..services.api_client import ApiClient

logger = logging.getLogger('ModelAdapter')

# Constants
PREDICT_EMBED_DEFAULT_SUBSET_LIMIT = 1000
PREDICT_EMBED_DEFAULT_TIMEOUT = 3600 * 24


class ModelConfigurations(MutableMapping):
    """
    Manages model configuration using composition with a backing dict.

    Uses MutableMapping to implement dict-like behavior without inheritance.
    This avoids duplication: if we inherited from dict, we'd have two dicts
    (one from inheritance, one from model_entity.configuration), leading to
    data inconsistency and maintenance issues.
    """

    def __init__(self, base_model_adapter):
        # Store reference to base_model_adapter dictionary
        self._backing_dict = {}

        if base_model_adapter is not None and base_model_adapter.model_entity is not None and base_model_adapter.model_entity.configuration is not None:
            self._backing_dict = base_model_adapter.model_entity.configuration
        if 'include_background' not in self._backing_dict:
            self._backing_dict['include_background'] = False
        self._base_model_adapter = base_model_adapter
        # Don't call _update_model_entity during initialization to avoid premature updates

    def _update_model_entity(self):
        if self._base_model_adapter is not None and self._base_model_adapter.model_entity is not None:
            self._base_model_adapter.model_entity.update(reload_services=False)

    def __ior__(self, other):
        self.update(other)
        return self

    # Required MutableMapping abstract methods
    def __getitem__(self, key):
        return self._backing_dict[key]

    def __setitem__(self, key, value):
        # Note: This method only updates the backing dict, not object attributes.
        # If you need to also update object attributes, be careful to avoid
        # infinite recursion by not calling __setattr__ from here.
        update = False
        if key not in self._backing_dict or self._backing_dict.get(key) != value:
            update = True
        self._backing_dict[key] = value
        if update:
            self._update_model_entity()

    def __delitem__(self, key):
        del self._backing_dict[key]

    def __iter__(self):
        return iter(self._backing_dict)

    def __len__(self):
        return len(self._backing_dict)

    def get(self, key, default=None):
        if key not in self._backing_dict:
            self.__setitem__(key, default)
        return self._backing_dict.get(key)

    def update(self, *args, **kwargs):
        # Check if there will be any modifications
        update_dict = dict(*args, **kwargs)
        has_changes = False
        for key, value in update_dict.items():
            if key not in self._backing_dict or self._backing_dict[key] != value:
                has_changes = True
                break
        self._backing_dict.update(*args, **kwargs)

        if has_changes:
            self._update_model_entity()

    def setdefault(self, key, default=None):
        if key not in self._backing_dict:
            self._backing_dict[key] = default
        return self._backing_dict[key]


@dataclasses.dataclass
class AdapterDefaults(ModelConfigurations):
    # for predict items, dataset, evaluate
    upload_annotations: bool = dataclasses.field(default=True)
    clean_annotations: bool = dataclasses.field(default=True)
    overwrite_annotations: bool = dataclasses.field(default=True)
    # for embeddings
    upload_features: bool = dataclasses.field(default=None)
    # for training
    root_path: str = dataclasses.field(default=None)
    data_path: str = dataclasses.field(default=None)
    output_path: str = dataclasses.field(default=None)

    def __init__(self, base_model_adapter=None):
        super().__init__(base_model_adapter)
        for f in dataclasses.fields(AdapterDefaults):
            # if the field exists in model_entity.configuration, use it
            # else set it from the attribute default value
            if super().get(f.name) is not None:
                super().__setattr__(f.name, super().get(f.name))
            else:
                super().__setitem__(f.name, f.default)

    def __setattr__(self, key, value):
        # Dataclass-like fields behave as attributes, so map to dict
        super().__setattr__(key, value)
        if not key.startswith("_"):
            super().__setitem__(key, value)

    def update(self, *args, **kwargs):
        for f in dataclasses.fields(AdapterDefaults):
            if f.name in kwargs:
                setattr(self, f.name, kwargs[f.name])
        super().update(*args, **kwargs)

    def resolve(self, key, *args):
        for arg in args:
            if arg is not None:
                super().__setitem__(key, arg)
                return arg
        return self.get(key, None)


class BaseModelAdapter(utilities.BaseServiceRunner):
    _client_api = attr.ib(type=ApiClient, repr=False)
    _feature_set_lock = threading.Lock()

    def __init__(self, model_entity: entities.Model = None):
        self.logger = logger
        # entities
        self._model_entity = None
        self._configuration = None
        self.adapter_defaults = None
        self.model = None
        self.bucket_path = None
        self._project = None
        self._feature_set = None
        # funcs
        self.item_to_batch_mapping = {'text': self._item_to_text, 'image': self._item_to_image}
        if model_entity is not None:
            self.load_from_model(model_entity=model_entity)
        logger.warning(
            "in case of a mismatch between 'model.name' and 'model_info.name' in the model adapter, model_info.name will be updated to align with 'model.name'."
        )

    ##################
    # Configurations #
    ##################

    @property
    def configuration(self) -> dict:
        # load from model
        if self._model_entity is not None:
            configuration = self._configuration
        else:
            configuration = dict()
        return configuration

    @configuration.setter
    def configuration(self, configuration: dict):
        assert isinstance(configuration, dict)
        if self._model_entity is not None:
            # Update configuration with received dict
            self._model_entity.configuration = configuration
            self.adapter_defaults = AdapterDefaults(self)
            self._configuration = self.adapter_defaults

    ############
    # Entities #
    ############
    @property
    def project(self):
        if self._project is None:
            self._project = self.model_entity.project
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def feature_set(self):
        if self._feature_set is None:
            self._feature_set = self._get_feature_set()
        assert isinstance(self._feature_set, entities.FeatureSet)
        return self._feature_set

    @property
    def model_entity(self):
        if self._model_entity is None:
            raise ValueError("No model entity loaded. Please load a model (adapter.load_from_model(<dl.Model>)) or set: 'adapter.model_entity=<dl.Model>'")
        assert isinstance(self._model_entity, entities.Model)
        return self._model_entity

    @model_entity.setter
    def model_entity(self, model_entity):
        assert isinstance(model_entity, entities.Model)
        if self._model_entity is not None and isinstance(self._model_entity, entities.Model):
            if self._model_entity.id != model_entity.id:
                self.logger.warning('Replacing Model from {!r} to {!r}'.format(self._model_entity.name, model_entity.name))
        self._model_entity = model_entity
        self.adapter_defaults = AdapterDefaults(self)
        self._configuration = self.adapter_defaults

    ###################################
    # NEED TO IMPLEMENT THESE METHODS #
    ###################################

    def load(self, local_path, **kwargs):
        """
        Loads model and populates self.model with a `runnable` model

        Virtual method - need to implement

        This function is called by load_from_model (download to local and then loads)

        :param local_path: `str` directory path in local FileSystem
        """
        raise NotImplementedError("Please implement `load` method in {}".format(self.__class__.__name__))

    def save(self, local_path, **kwargs):
        """
        Saves configuration and weights locally

        Virtual method - need to implement

        the function is called in save_to_model which first save locally and then uploads to model entity

        :param local_path: `str` directory path in local FileSystem
        """
        raise NotImplementedError("Please implement `save` method in {}".format(self.__class__.__name__))

    def train(self, data_path, output_path, **kwargs):
        """
        Virtual method - need to implement

        Train the model according to data in data_paths and save the train outputs to output_path,
        this include the weights and any other artifacts created during train

        :param data_path: `str` local File System path to where the data was downloaded and converted at
        :param output_path: `str` local File System path where to dump training mid-results (checkpoints, logs...)
        """
        raise NotImplementedError("Please implement `train` method in {}".format(self.__class__.__name__))

    def predict(self, batch, **kwargs):
        """
        Model inference (predictions) on batch of items

        Virtual method - need to implement

        :param batch: output of the `prepare_item_func` func
        :return: `list[dl.AnnotationCollection]` each collection is per each image / item in the batch
        """
        raise NotImplementedError("Please implement `predict` method in {}".format(self.__class__.__name__))

    def embed(self, batch, **kwargs):
        """
        Extract model embeddings on batch of items

        Virtual method - need to implement

        :param batch: output of the `prepare_item_func` func
        :return: `list[list]` a feature vector per each item in the batch
        """
        raise NotImplementedError("Please implement `embed` method in {}".format(self.__class__.__name__))

    def evaluate(self, model: entities.Model, dataset: entities.Dataset, filters: entities.Filters) -> entities.Model:
        """
        This function evaluates the model prediction on a dataset (with GT annotations).
        The evaluation process will upload the scores and metrics to the platform.

        :param model: The model to evaluate (annotation.metadata.system.model.name
        :param dataset: Dataset where the model predicted and uploaded its annotations
        :param filters: Filters to query items on the dataset
        :return:
        """
        import dtlpymetrics

        compare_types = model.output_type
        if not filters:
            filters = entities.Filters()
        if filters is not None and isinstance(filters, dict):
            filters = entities.Filters(custom_filter=filters)
        model = dtlpymetrics.scoring.create_model_score(
            model=model,
            dataset=dataset,
            filters=filters,
            compare_types=compare_types,
        )
        return model

    def convert_from_dtlpy(self, data_path, **kwargs):
        """Convert Dataloop structure data to model structured

            Virtual method - need to implement

            e.g. take dlp dir structure and construct annotation file

        :param data_path: `str` local File System directory path where we already downloaded the data from dataloop platform
        :return:
        """
        raise NotImplementedError("Please implement `convert_from_dtlpy` method in {}".format(self.__class__.__name__))

    #################
    # DTLPY METHODS #
    ################
    def prepare_item_func(self, item: entities.Item):
        """
        Prepare the Dataloop item before calling the `predict` function with a batch.
        A user can override this function to load item differently
        Default will load the item according the input_type (mapping type to function is in self.item_to_batch_mapping)

        :param item:
        :return: preprocessed: the var with the loaded item information (e.g. ndarray for image, dict for json files etc)
        """
        # Item to batch func
        if isinstance(self.model_entity.input_type, list):
            if 'text' in self.model_entity.input_type and 'text' in item.mimetype:
                processed = self._item_to_text(item)
            elif 'image' in self.model_entity.input_type and 'image' in item.mimetype:
                processed = self._item_to_image(item)
            else:
                processed = self._item_to_item(item)

        elif self.model_entity.input_type in self.item_to_batch_mapping:
            processed = self.item_to_batch_mapping[self.model_entity.input_type](item)

        else:
            processed = self._item_to_item(item)

        return processed

    def __include_model_annotations(self, annotation_filters):
        include_model_annotations = self.model_entity.configuration.get("include_model_annotations", False)
        if include_model_annotations is False:
            if annotation_filters.custom_filter is None:
                annotation_filters.add(field="metadata.system.model.name", values=False, operator=entities.FiltersOperations.EXISTS)
            else:
                annotation_filters.custom_filter['filter']['$and'].append({'metadata.system.model.name': {'$exists': False}})
        return annotation_filters

    def __download_items(self, dataset, filters, local_path, annotation_options, annotation_filters=None):
        """
        Download items from dataset with optional annotation filters.

        :param dataset: Dataset to download from
        :param filters: Filters to apply
        :param local_path: Local path to save files
        :param annotation_options: Annotation download options
        :param annotation_filters: Optional filters for annotations
        :return: List of downloaded items
        """
        if annotation_options == entities.ViewAnnotationOptions.JSON:
            downloader = repositories.Downloader(dataset.items)
            return downloader._download_recursive(
                local_path=local_path,
                filters=filters,
                annotation_filters=annotation_filters
            )
        else:
            return dataset.items.download(
                filters=filters,
                local_path=local_path,
                annotation_options=annotation_options,
                annotation_filters=annotation_filters
            )

    def __download_background_images(self, filters, data_subset_base_path, annotation_options):
        background_list = list()
        if self.configuration.get('include_background', False) is True:
            filters.custom_filter["filter"]["$and"].append({"annotated": False})
            background_list = self.__download_items(
                dataset=self.model_entity.dataset,
                filters=filters,
                local_path=data_subset_base_path,
                annotation_options=annotation_options
            )
        return background_list

    def prepare_data(
        self,
        dataset: entities.Dataset,
        # paths
        root_path=None,
        data_path=None,
        output_path=None,
        #
        overwrite=False,
        **kwargs,
    ):
        """
        Prepares dataset locally before training or evaluation.
        download the specific subset selected to data_path and preforms `self.convert` to the data_path dir

        :param dataset: dl.Dataset
        :param root_path: `str` root directory for training. default is "tmp". Can be set using self.adapter_defaults.root_path
        :param data_path: `str` dataset directory. default <root_path>/"data". Can be set using self.adapter_defaults.data_path
        :param output_path: `str` save everything to this folder. default <root_path>/"output". Can be set using self.adapter_defaults.output_path

        :param bool overwrite: overwrite the data path (download again). default is False
        """
        # define paths
        dataloop_path = service_defaults.DATALOOP_PATH
        root_path = self.adapter_defaults.resolve("root_path", root_path)
        data_path = self.adapter_defaults.resolve("data_path", data_path)
        output_path = self.adapter_defaults.resolve("output_path", output_path)
        if root_path is None:
            now = datetime.datetime.now()
            root_path = os.path.join(
                dataloop_path,
                'model_data',
                "{s_id}_{s_n}".format(s_id=self.model_entity.id, s_n=self.model_entity.name),
                now.strftime('%Y-%m-%d-%H%M%S'),
            )
        if data_path is None:
            data_path = os.path.join(root_path, 'datasets', self.model_entity.dataset.id)
            os.makedirs(data_path, exist_ok=True)
        if output_path is None:
            output_path = os.path.join(root_path, 'output')
            os.makedirs(output_path, exist_ok=True)

        if len(os.listdir(data_path)) > 0:
            self.logger.warning("Data path directory ({}) is not empty..".format(data_path))

        annotation_options = entities.ViewAnnotationOptions.JSON
        if self.model_entity.output_type in [entities.AnnotationType.SEGMENTATION]:
            annotation_options = entities.ViewAnnotationOptions.INSTANCE

        # Download the subset items
        subsets = self.model_entity.metadata.get("system", {}).get("subsets", None)
        annotations_subsets = self.model_entity.metadata.get("system", {}).get("annotationsSubsets", {})
        if subsets is None:
            raise ValueError("Model (id: {}) must have subsets in metadata.system.subsets".format(self.model_entity.id))
        for subset, filters_dict in subsets.items():
            _filters_dict = filters_dict.copy()
            data_subset_base_path = os.path.join(data_path, subset)
            if os.path.isdir(data_subset_base_path) and not overwrite:
                # existing and dont overwrite
                self.logger.debug("Subset {!r} already exists (and overwrite=False). Skipping.".format(subset))
                continue

            filters = entities.Filters(custom_filter=_filters_dict)
            self.logger.debug("Downloading subset {!r} of {}".format(subset, self.model_entity.dataset.name))

            annotation_filters = None
            if self.model_entity.output_type != "embedding":
                if subset in annotations_subsets:
                    annotation_filters = entities.Filters(
                        use_defaults=False,
                        resource=entities.FiltersResource.ANNOTATION,
                        custom_filter=annotations_subsets[subset],
                    )
                # if user provided annotation_filters, skip the default filters
                elif self.model_entity.output_type is not None:
                    annotation_filters = entities.Filters(resource=entities.FiltersResource.ANNOTATION, use_defaults=False)
                    if self.model_entity.output_type in [
                        entities.AnnotationType.SEGMENTATION,
                        entities.AnnotationType.POLYGON,
                    ]:
                        model_output_types = [entities.AnnotationType.SEGMENTATION, entities.AnnotationType.POLYGON]
                    else:
                        model_output_types = [self.model_entity.output_type]

                    annotation_filters.add(
                        field=entities.FiltersKnownFields.TYPE,
                        values=model_output_types,
                        operator=entities.FiltersOperations.IN,
                    )

                annotation_filters = self.__include_model_annotations(annotation_filters)
                annotations_subsets[subset] = annotation_filters.prepare()

            ret_list = self.__download_items(
                dataset=dataset,
                filters=filters,
                local_path=data_subset_base_path,
                annotation_options=annotation_options,
                annotation_filters=annotation_filters
            )
            _filters_dict = subsets[subset].copy()
            filters = entities.Filters(custom_filter=_filters_dict)
            background_ret_list = self.__download_background_images(
                filters=filters,
                data_subset_base_path=data_subset_base_path,
                annotation_options=annotation_options,
            )
            ret_list = list(ret_list)
            background_ret_list = list(background_ret_list)
            self.logger.debug(f"Subset '{subset}': ret_list length: {len(ret_list)}, background_ret_list length: {len(background_ret_list)}")
            # Combine ret_list and background_ret_list generators into a single generator
            ret_list = ret_list + background_ret_list
            if isinstance(ret_list, list) and len(ret_list) == 0:
                if annotation_filters is not None:
                    annotation_filters_str = annotation_filters.prepare()
                else:
                    annotation_filters_str = None
                raise ValueError(
                    f"No items downloaded for subset {subset}! Cannot train model with empty subset.\n"
                    f"Subset {subset} filters: {filters.prepare()}\n"
                    f"Annotation filters: {annotation_filters_str}"
                )

        self.convert_from_dtlpy(data_path=data_path, **kwargs)
        return root_path, data_path, output_path

    def load_from_model(self, model_entity=None, local_path=None, overwrite=True, **kwargs):
        """Loads a model from given `dl.Model`.
            Reads configurations and instantiate self.model_entity
            Downloads the model_entity bucket (if available)

        :param model_entity:  `str` dl.Model entity
        :param local_path:  `str` directory path in local FileSystem to download the model_entity to
        :param overwrite: `bool` (default False) if False does not download files with same name else (True) download all
        """
        if model_entity is not None:
            self.model_entity = model_entity
        if local_path is None:
            local_path = os.path.join(service_defaults.DATALOOP_PATH, "models", self.model_entity.name)
        # Load configuration and adapter defaults
        self.adapter_defaults = AdapterDefaults(self)
        # Point _configuration to the same object since AdapterDefaults inherits from ModelConfigurations
        self._configuration = self.adapter_defaults
        # Download
        self.model_entity.artifacts.download(local_path=local_path, overwrite=overwrite)
        self.load(local_path, **kwargs)

    def save_to_model(self, local_path=None, cleanup=False, replace=True, **kwargs):
        """
        Saves the model state to a new bucket and configuration

        Saves configuration and weights to artifacts
        Mark the model as `trained`
        loads only applies for remote buckets

        :param local_path: `str` directory path in local FileSystem to save the current model bucket (weights) (default will create a temp dir)
        :param replace: `bool` will clean the bucket's content before uploading new files
        :param cleanup: `bool` if True (default) remove the data from local FileSystem after upload
        :return:
        """

        if local_path is None:
            local_path = tempfile.mkdtemp(prefix="model_{}".format(self.model_entity.name))
            self.logger.debug("Using temporary dir at {}".format(local_path))

        self.save(local_path=local_path, **kwargs)

        if self.model_entity is None:
            raise ValueError('Missing model entity on the adapter. ' 'Please set before saving: "adapter.model_entity=model"')

        self.model_entity.artifacts.upload(filepath=os.path.join(local_path, '*'), overwrite=True)
        if cleanup:
            shutil.rmtree(path=local_path, ignore_errors=True)
            self.logger.info("Clean-up. deleting {}".format(local_path))

    # ===============
    # SERVICE METHODS
    # ===============

    @entities.Package.decorators.function(
        display_name='Predict Items',
        inputs={'items': 'Item[]'},
        outputs={'items': 'Item[]', 'annotations': 'Annotation[]'},
    )
    def predict_items(self, items: list, batch_size=None, **kwargs):
        """
        Run the predict function on the input list of items (or single) and return the items and the predictions.
        Each prediction is by the model output type (package.output_type) and model_info in the metadata

        :param items: `List[dl.Item]` list of items to predict
        :param batch_size: `int` size of batch to run a single inference

        :return: `List[dl.Item]`, `List[List[dl.Annotation]]`
        """
        if batch_size is None:
            batch_size = self.configuration.get('batch_size', 4)
        input_type = self.model_entity.input_type
        self.logger.debug("Predicting {} items, using batch size {}. input type: {}".format(len(items), batch_size, input_type))
        pool = ThreadPoolExecutor(max_workers=16)
        error_counter = 0
        fail_ids = list()
        annotations = list()
        for i_batch in tqdm.tqdm(range(0, len(items), batch_size), desc='predicting', unit='bt', leave=None, file=sys.stdout):
            batch_items = items[i_batch : i_batch + batch_size]
            batch = list(pool.map(self.prepare_item_func, batch_items))
            try:
                batch_collections = self.predict(batch, **kwargs)
            except Exception as e:
                item_ids = [item.id for item in batch_items]
                self.logger.error(f"Failed to predict batch {i_batch} for items {item_ids}. Error: {e}\n{traceback.format_exc()}")
                error_counter += 1
                fail_ids.extend(item_ids)
                continue
            _futures = list(pool.map(partial(self._update_predictions_metadata), batch_items, batch_collections))
            # Loop over the futures to make sure they are all done to avoid race conditions
            _ = [_f for _f in _futures]
            self.logger.debug("Uploading items' annotation for model {!r}.".format(self.model_entity.name))
            try:
                batch_collections = list(
                    pool.map(partial(self._upload_model_annotations), batch_items, batch_collections)
                )
            except Exception as err:
                item_ids = [item.id for item in batch_items]
                self.logger.error(
                    f"Failed to upload annotations for items {item_ids}. Error: {err}\n{traceback.format_exc()}"
                )
                error_counter += 1
                fail_ids.extend(item_ids)

            for collection in batch_collections:
                # function needs to return `List[List[dl.Annotation]]`
                # convert annotation collection to a list of dl.Annotation for each batch
                if isinstance(collection, entities.AnnotationCollection):
                    annotations.extend([annotation for annotation in collection.annotations])
                else:
                    logger.warning(f'RETURN TYPE MAY BE INVALID: {type(collection)}')
                    annotations.extend(collection)
            # TODO call the callback

        pool.shutdown()
        if error_counter > 0:
            raise Exception(f"Failed to predict all items. Failed IDs: {fail_ids}, See logs for more details")
        return items, annotations

    @entities.Package.decorators.function(
        display_name='Embed Items',
        inputs={'items': 'Item[]'},
        outputs={'items': 'Item[]', 'features': 'Json[]'},
    )
    def embed_items(self, items: list, upload_features=None, batch_size=None, progress: utilities.Progress = None, **kwargs):
        """
        Extract feature from an input list of items (or single) and return the items and the feature vector.

        :param items: `List[dl.Item]` list of items to embed
        :param upload_features: `bool` uploads the features on the given items
        :param batch_size: `int` size of batch to run a single embed

        :return: `List[dl.Item]`, `List[List[vector]]`
        """
        if batch_size is None:
            batch_size = self.configuration.get('batch_size', 4)
        upload_features = self.adapter_defaults.resolve("upload_features", upload_features)
        skip_default_items = upload_features is None
        if upload_features is None:
            upload_features = True
        input_type = self.model_entity.input_type
        self.logger.debug("Embedding {} items, using batch size {}. input type: {}".format(len(items), batch_size, input_type))
        error_counter = 0
        fail_ids = list()

        feature_set = self.feature_set

        # upload the feature vectors
        pool = ThreadPoolExecutor(max_workers=16)
        vectors = list()
        _items = list()
        for i_batch in tqdm.tqdm(
            range(0, len(items), batch_size),
            desc='embedding',
            unit='bt',
            leave=None,
            file=sys.stdout,
        ):
            batch_items = items[i_batch : i_batch + batch_size]
            batch = list(pool.map(self.prepare_item_func, batch_items))
            try:
                batch_vectors = self.embed(batch, **kwargs)
            except Exception as err:
                item_ids = [item.id for item in batch_items]
                self.logger.error(f"Failed to embed batch {i_batch} for items {item_ids}. Error: {err}\n{traceback.format_exc()}")
                error_counter += 1
                fail_ids.extend(item_ids)
                continue
            vectors.extend(batch_vectors)
            # Save the items in the order of the vectors
            _items.extend(batch_items)
        pool.shutdown()

        if upload_features is True:
            embeddings_size = self.configuration.get('embeddings_size', 256)
            valid_items = []
            valid_vectors = []
            items_to_upload = []
            vectors_to_upload = []

            for item, vector in zip(_items, vectors):
                # Check if vector is valid
                if vector is None or len(vector) != embeddings_size:
                    self.logger.warning(f"Vector generated for item {item.id} is None or has wrong size. Skipping...")
                    continue

                # Item and vector are valid
                valid_items.append(item)
                valid_vectors.append(vector)

                # Check if item should be skipped (prompt items)
                _system_metadata = getattr(item, 'system', dict())
                is_prompt = _system_metadata.get('shebang', dict()).get('dltype', '') == 'prompt'
                if skip_default_items and is_prompt:
                    self.logger.debug(f"Skipping feature upload for prompt item {item.id}")
                    continue

                # Items were not skipped - should be uploaded
                items_to_upload.append(item)
                vectors_to_upload.append(vector)

            # Update the original lists with valid items only
            _items[:] = valid_items
            vectors[:] = valid_vectors

            if len(_items) != len(vectors):
                raise ValueError(f"The number of items ({len(_items)}) is not equal to the number of vectors ({len(vectors)}).")

            self.logger.debug(f"Uploading {len(items_to_upload)} items' feature vectors for model {self.model_entity.name}.")
            try:
                start_time = time.time()
                feature_set.features.create(entity=items_to_upload, value=vectors_to_upload, feature_set_id=feature_set.id, project_id=self.model_entity.project_id)
                self.logger.debug(f"Uploaded {len(items_to_upload)} items' feature vectors for model {self.model_entity.name} in {time.time() - start_time} seconds.")
            except Exception as err:
                self.logger.error(f"Failed to upload feature vectors. Error: {err}\n{traceback.format_exc()}")
                error_counter += 1
        if error_counter > 0:
            raise Exception(f"Failed to embed all items. Failed IDs: {fail_ids}, See logs for more details")
        return _items, vectors

    @entities.Package.decorators.function(
        display_name='Embed Dataset with DQL',
        inputs={'dataset': 'Dataset', 'filters': 'Json'},
    )
    def embed_dataset(
        self,
        dataset: entities.Dataset,
        filters: Optional[entities.Filters] = None,
        upload_features: Optional[bool] = None,
        batch_size: Optional[int] = None,
        progress: Optional[utilities.Progress] = None,
        **kwargs,
    ):
        """
        Run model embedding on all items in a dataset

        :param dataset: Dataset entity to embed
        :param filters: Filters entity for filtering before embedding
        :param upload_features: bool whether to upload features back to platform
        :param batch_size: int size of batch to run a single embedding
        :param progress: dl.Progress object to track progress
        :return: bool indicating if embedding completed successfully
        """

        self._execute_dataset_operation(
            dataset=dataset,
            operation_type='embed',
            filters=filters,
            progress=progress,
            batch_size=batch_size,
        )

    @entities.Package.decorators.function(
        display_name='Predict Dataset with DQL',
        inputs={'dataset': 'Dataset', 'filters': 'Json'},
    )
    def predict_dataset(
        self,
        dataset: entities.Dataset,
        filters: Optional[entities.Filters] = None,
        batch_size: Optional[int] = None,
        progress: Optional[utilities.Progress] = None,
        **kwargs,
    ):
        """
        Run model prediction on all items in a dataset

        :param dataset: Dataset entity to predict
        :param filters: Filters entity for filtering before prediction
        :param batch_size: int size of batch to run a single prediction
        :param progress: dl.Progress object to track progress
        :return: bool indicating if prediction completed successfully
        """
        self._execute_dataset_operation(
            dataset=dataset,
            operation_type='predict',
            filters=filters,
            progress=progress,
            batch_size=batch_size,
        )

    @entities.Package.decorators.function(
        display_name='Train a Model',
        inputs={'model': entities.Model},
        outputs={'model': entities.Model},
    )
    def train_model(self, model: entities.Model, cleanup=False, progress: utilities.Progress = None, context: utilities.Context = None):
        """
        Train on existing model.
        data will be taken from dl.Model.datasetId
        configuration is as defined in dl.Model.configuration
        upload the output the model's bucket (model.bucket)
        """
        if isinstance(model, dict):
            model = repositories.Models(client_api=self._client_api).get(model_id=model['id'])
        output_path = None
        try:
            logger.info("Received {s} for training".format(s=model.id))
            model = model.wait_for_model_ready()
            if model.status == 'failed':
                logger.warning("Model failed. New training will attempt to resume from previous checkpoints.")

            ##############
            # Set status #
            ##############
            model.status = 'training'
            if context is not None:
                if 'system' not in model.metadata:
                    model.metadata['system'] = dict()
            model.update(reload_services=False)

            ##########################
            # load model and weights #
            ##########################
            logger.info("Loading Adapter with: {n} ({i!r})".format(n=model.name, i=model.id))
            self.load_from_model(model_entity=model)

            ################
            # prepare data #
            ################
            root_path, data_path, output_path = self.prepare_data(dataset=self.model_entity.dataset, root_path=os.path.join('tmp', model.id))
            # Start the Train
            logger.info(f"Training model {model.name!r} ({model.id!r}) on data {data_path!r}")
            if progress is not None:
                progress.update(message='starting training')

            def on_epoch_end_callback(i_epoch, n_epoch):
                if progress is not None:
                    progress.update(progress=int(100 * (i_epoch + 1) / n_epoch), message='finished epoch: {}/{}'.format(i_epoch, n_epoch))

            self.train(data_path=data_path, output_path=output_path, on_epoch_end_callback=on_epoch_end_callback)
            if progress is not None:
                progress.update(message='saving model', progress=99)

            self.save_to_model(local_path=output_path, replace=True)
            model.status = 'trained'
            model.update(reload_services=False)
            ###########
            # cleanup #
            ###########
            if cleanup:
                shutil.rmtree(output_path, ignore_errors=True)
        except Exception:
            # save also on fail
            if output_path is not None:
                self.save_to_model(local_path=output_path, replace=True)
            logger.info('Execution failed. Setting model.status to failed')
            raise
        return model

    @entities.Package.decorators.function(
        display_name='Evaluate a Model',
        inputs={'model': entities.Model, 'dataset': entities.Dataset, 'filters': 'Json'},
        outputs={'model': entities.Model, 'dataset': entities.Dataset},
    )
    def evaluate_model(
        self,
        model: entities.Model,
        dataset: entities.Dataset,
        filters: entities.Filters = None,
        #
        progress: utilities.Progress = None,
        context: utilities.Context = None,
    ):
        """
        Evaluate a model.
        data will be downloaded from the dataset and query
        configuration is as defined in dl.Model.configuration
        upload annotations and calculate metrics vs GT

        :param model: Model entity to run prediction
        :param dataset: Dataset to evaluate
        :param filters: Filter for specific items from dataset
        :param progress: dl.Progress for report FaaS progress
        :param context:
        :return:
        """
        logger.info(f"Received model: {model.id} for evaluation on dataset (name: {dataset.name}, id: {dataset.id}")
        ##########################
        # load model and weights #
        ##########################
        logger.info(f"Loading Adapter with: {model.name} ({model.id!r})")
        self.load_from_model(dataset=dataset, model_entity=model)

        ##############
        # Predicting #
        ##############
        logger.info(f"Calling prediction, dataset: {dataset.name!r} ({model.id!r}), filters: {filters}")
        if not filters:
            filters = entities.Filters()
        if self.adapter_defaults.get("overwrite_annotations", True) is True:
            self._execute_dataset_operation(
                dataset=dataset,
                operation_type='predict',
                filters=filters,
                multiple_executions=False,
            )

        ##############
        # Evaluating #
        ##############
        logger.info(f"Starting adapter.evaluate()")
        if progress is not None:
            progress.update(message='calculating metrics', progress=98)
        model = self.evaluate(model=model, dataset=dataset, filters=filters)
        #########
        # Done! #
        #########
        if progress is not None:
            progress.update(message='finishing evaluation', progress=99)
        return model, dataset

    # =============
    # INNER METHODS
    # =============
    def _get_feature_set(self):
        # Ensure feature set creation/retrieval is thread-safe across the class
        with self.__class__._feature_set_lock:
            # Search for existing feature set for this model id
            feature_set = self.model_entity.feature_set
            if feature_set is None:
                logger.info('Feature Set not found. creating... ')
                try:
                    self.project.feature_sets.get(feature_set_name=self.model_entity.name)
                    feature_set_name = f"{self.model_entity.name}-{''.join(random.choices(string.ascii_letters + string.digits, k=5))}"
                    logger.warning(
                        f"Feature set with the model name already exists. Creating new feature set with name {feature_set_name}"
                    )

                except exceptions.NotFound:
                    feature_set_name = self.model_entity.name
                feature_set = self.project.feature_sets.create(
                    name=feature_set_name,
                    entity_type=entities.FeatureEntityType.ITEM,
                    model_id=self.model_entity.id,
                    project_id=self.project.id,
                    set_type=self.model_entity.name,
                    size=self.configuration.get('embeddings_size', 256),
                )
                logger.info(f'Feature Set created! name: {feature_set.name}, id: {feature_set.id}')
            else:
                logger.info(f'Feature Set found! name: {feature_set.name}, id: {feature_set.id}')
            return feature_set

    def _execute_dataset_operation(
        self,
        dataset: entities.Dataset,
        operation_type: str,
        filters: Optional[entities.Filters] = None,
        progress: Optional[utilities.Progress] = None,
        batch_size: Optional[int] = None,
        multiple_executions: bool = True,
    ) -> bool:
        """
        Execute dataset operation (predict/embed) with batching and filtering support.

        :param dataset: Dataset entity to run operation on
        :param operation_type: Type of operation to execute ('predict' or 'embed')
        :param filters: Filters entity to filter items, default None
        :param progress: Progress object for tracking progress, default None
        :param batch_size: Size of batches to process items, default None (uses model config)
        :param multiple_executions: Whether to use multiple executions when filters exceed subset limit, default True
        :return: True if operation completes successfully
        :raises ValueError: If operation_type is not 'predict' or 'embed'
        """
        self.logger.debug(f"Running {operation_type} for dataset (name:{dataset.name}, id:{dataset.id})")

        if not filters:
            self.logger.debug("No filters provided, using default filters")
            filters = entities.Filters()
        if filters is not None and isinstance(filters, dict):
            self.logger.debug(f"Received custom filters {filters}")
            filters = entities.Filters(custom_filter=filters)

        if operation_type == 'embed':
            feature_set = self.feature_set
            logger.info(f"Feature set found! name: {feature_set.name}, id: {feature_set.id}")

        predict_embed_subset_limit = self.configuration.get('predict_embed_subset_limit', PREDICT_EMBED_DEFAULT_SUBSET_LIMIT)
        predict_embed_timeout = self.configuration.get('predict_embed_timeout', PREDICT_EMBED_DEFAULT_TIMEOUT)
        self.logger.debug(f"Inputs: predict_embed_subset_limit: {predict_embed_subset_limit}, predict_embed_timeout: {predict_embed_timeout}")
        tmp_filters = copy.deepcopy(filters.prepare())
        tmp_filters['pageSize'] = 0
        num_items = dataset.items.list(filters=entities.Filters(custom_filter=tmp_filters)).items_count
        self.logger.debug(f"Number of items for current filters: {num_items}")

        # One-item lookahead on generator: if only one subset, run locally; else create executions for all
        gen = entities.Filters._get_split_filters(dataset, filters, predict_embed_subset_limit)
        try:
            first_filter = next(gen)
        except StopIteration:
            self.logger.info("Filters is empty, nothing to run")
            return True

        try:
            second_filter = next(gen)
            multiple = True
        except StopIteration:
            multiple = False

        # Create consistent iterable of all filters for reuse
        # Both paths use chain to ensure consistent type and iteration behavior
        if multiple:
            # Chain together the pre-consumed filters with the remaining generator
            all_filters = chain([first_filter, second_filter], gen)
        else:
            # Single filter - use chain with empty generator for consistency
            all_filters = chain([first_filter], [])

        if not multiple or not multiple_executions:
            self.logger.info("Running locally")
            if batch_size is None:
                batch_size = self.configuration.get('batch_size', 4)

            # Process each filter locally
            for filter_dict in all_filters:
                filter_dict["pageSize"] = 1000
                single_filters = entities.Filters(custom_filter=filter_dict)
                pages = dataset.items.list(filters=single_filters)
                self.logger.info(f"Processing filter on: {pages.items_count} items")
                items = [item for page in pages for item in page if item.type == 'file']
                self.logger.debug(f"Items length: {len(items)}")

                if operation_type == 'embed':
                    self.embed_items(items=items, batch_size=batch_size, progress=progress)
                elif operation_type == 'predict':
                    self.predict_items(items=items, batch_size=batch_size, progress=progress)
                else:
                    raise ValueError(f"Unsupported operation type: {operation_type}")
            return True

        executions = []
        for filter_dict in all_filters:
            self.logger.debug(f"Creating execution for models {operation_type} with dataset id {dataset.id} and filter_dict {filter_dict}")
            if operation_type == 'embed':
                execution = self.model_entity.models.embed(
                    model=self.model_entity,
                    dataset_id=dataset.id,
                    filters=entities.Filters(custom_filter=filter_dict),
                )
            elif operation_type == 'predict':
                execution = self.model_entity.models.predict(
                    model=self.model_entity, dataset_id=dataset.id, filters=entities.Filters(custom_filter=filter_dict)
                )
            else:
                raise ValueError(f"Unsupported operation type: {operation_type}")
            executions.append(execution)

        if executions:
            self.logger.info(f'Created {len(executions)} executions for {operation_type}, ' f'execution ids: {[ex.id for ex in executions]}')

            wait_time = 5
            start_time = time.time()
            last_perc = 0
            self.logger.debug(f"Waiting for executions with timeout {predict_embed_timeout}")
            while time.time() - start_time < predict_embed_timeout:
                continue_loop = False
                total_perc = 0

                for ex in executions:
                    execution = self.project.executions.get(execution_id=ex.id)
                    perc = execution.latest_status.get('percentComplete', 0)
                    total_perc += perc
                    if execution.in_progress():
                        continue_loop = True

                avg_perc = round(total_perc / len(executions), 0)
                if progress is not None and last_perc != avg_perc:
                    last_perc = avg_perc
                    progress.update(progress=last_perc, message=f'running {operation_type}')

                if not continue_loop:
                    break

                time.sleep(wait_time)
            self.logger.debug("End waiting for executions")
            # Check if any execution failed
            executions_filter = entities.Filters(resource=entities.FiltersResource.EXECUTION)
            executions_filter.add(field="id", values=[ex.id for ex in executions], operator=entities.FiltersOperations.IN)
            executions_filter.add(field='latestStatus.status', values='failed')
            executions_filter.page_size = 0
            failed_executions_count = self.project.executions.list(filters=executions_filter).items_count
            if failed_executions_count > 0:
                self.logger.error(f"Failed to {operation_type} for {failed_executions_count} executions")
                raise ValueError(f"Failed to {operation_type} entire dataset, please check the logs for more details")
        return True

    def _upload_model_annotations(self, item: entities.Item, predictions):
        """
        Utility function that upload prediction to dlp platform based on the package.output_type
        :param predictions: `dl.AnnotationCollection`
        :param cleanup: `bool` if set removes existing predictions with the same package-model name
        """
        if not (isinstance(predictions, entities.AnnotationCollection) or isinstance(predictions, list)):
            raise TypeError(f'predictions was expected to be of type {entities.AnnotationCollection}, but instead it is {type(predictions)}')
        clean_filter = entities.Filters(resource=entities.FiltersResource.ANNOTATION)
        clean_filter.add(field='metadata.user.model.name', values=self.model_entity.name, method=entities.FiltersMethod.OR)
        clean_filter.add(field='metadata.system.model.name', values=self.model_entity.name, method=entities.FiltersMethod.OR)
        # clean_filter.add(field='type', values=self.model_entity.output_type,)
        item.annotations.delete(filters=clean_filter)
        annotations = item.annotations.upload(annotations=predictions)
        return annotations

    @staticmethod
    def _item_to_image(item):
        """
        Preprocess items before calling the `predict` functions.
        Convert item to numpy array

        :param item:
        :return:
        """
        try:
            buffer = item.download(save_locally=False)
            image = np.asarray(Image.open(buffer))
        except Exception as e:
            logger.error(f"Failed to convert image to np.array, Error: {e}\n{traceback.format_exc()}")
            image = None
        return image

    @staticmethod
    def _item_to_item(item):
        """
        Default item to batch function.
        This function should prepare a single item for the predict function, e.g. for images, it loads the image as numpy array
        :param item:
        :return:
        """
        return item

    @staticmethod
    def _item_to_text(item):
        filename = item.download(overwrite=True)
        text = None
        if item.mimetype == 'text/plain' or item.mimetype == 'text/markdown':
            with open(filename, 'r') as f:
                text = f.read()
                text = text.replace('\n', ' ')
        else:
            logger.warning('Item is not text file. mimetype: {}'.format(item.mimetype))
            text = item
        if os.path.exists(filename):
            os.remove(filename)
        return text

    @staticmethod
    def _uri_to_image(data_uri):
        # data_uri = "data:image/png;base64, "
        image_b64 = data_uri.split(",")[1]
        binary = base64.b64decode(image_b64)
        image = np.asarray(Image.open(io.BytesIO(binary)))
        return image

    def _update_predictions_metadata(self, item: entities.Item, predictions: entities.AnnotationCollection):
        """
        add model_name and model_id to the metadata of the annotations.
        add model_info to the metadata of the system metadata of the annotation.
        Add item id to all the annotations in the AnnotationCollection

        :param item: Entity.Item
        :param predictions: item's AnnotationCollection
        :return:
        """
        for prediction in predictions:
            if prediction.type == entities.AnnotationType.SEGMENTATION:
                color = None
                try:
                    color = item.dataset._get_ontology().color_map.get(prediction.label, None)
                except (exceptions.BadRequest, exceptions.NotFound):
                    ...
                if color is None:
                    if self.model_entity._dataset is not None:
                        try:
                            color = self.model_entity.dataset._get_ontology().color_map.get(prediction.label, (255, 255, 255))
                        except (exceptions.BadRequest, exceptions.NotFound):
                            ...
                if color is None:
                    logger.warning("Can't get annotation color from model's dataset, using default.")
                    color = prediction.color
                prediction.color = color

            prediction.item_id = item.id
            if 'user' in prediction.metadata and 'model' in prediction.metadata['user']:
                prediction.metadata['user']['model']['model_id'] = self.model_entity.id
                prediction.metadata['user']['model']['name'] = self.model_entity.name
            if 'system' not in prediction.metadata:
                prediction.metadata['system'] = dict()
            if 'model' not in prediction.metadata['system']:
                prediction.metadata['system']['model'] = dict()
            confidence = prediction.metadata.get('user', dict()).get('model', dict()).get('confidence', None)
            prediction.metadata['system']['model'] = {
                'model_id': self.model_entity.id,
                'name': self.model_entity.name,
                'confidence': confidence,
            }

    ##############################
    # Callback Factory functions #
    ##############################
    @property
    def dataloop_keras_callback(self):
        """
        Returns the constructor for a keras api dump callback
        The callback is used for dlp platform to show train losses

        :return: DumpHistoryCallback constructor
        """
        try:
            import keras
        except (ImportError, ModuleNotFoundError) as err:
            raise RuntimeError(f'{self.__class__.__name__} depends on extenral package. Please install ') from err

        import os
        import time
        import json

        class DumpHistoryCallback(keras.callbacks.Callback):
            def __init__(self, dump_path):
                super().__init__()
                if os.path.isdir(dump_path):
                    dump_path = os.path.join(dump_path, f'__view__training-history__{time.strftime("%F-%X")}.json')
                self.dump_file = dump_path
                self.data = dict()

            def on_epoch_end(self, epoch, logs=None):
                logs = logs or {}
                for name, val in logs.items():
                    if name not in self.data:
                        self.data[name] = {'x': list(), 'y': list()}
                    self.data[name]['x'].append(float(epoch))
                    self.data[name]['y'].append(float(val))
                self.dump_history()

            def dump_history(self):
                _json = {
                    "query": {},
                    "datasetId": "",
                    "xlabel": "epoch",
                    "title": "training loss",
                    "ylabel": "val",
                    "type": "metric",
                    "data": [
                        {
                            "name": name,
                            "x": values['x'],
                            "y": values['y'],
                        }
                        for name, values in self.data.items()
                    ],
                }

                with open(self.dump_file, 'w') as f:
                    json.dump(_json, f, indent=2)

        return DumpHistoryCallback



from . import BaseAnnotationDefinition


class Description(BaseAnnotationDefinition):
    """
        Subtitle annotation object
    """

    def __init__(self, text, description=None):
        super().__init__(description=description)
        self.type = "item_description"
        self.text = text
        self.label = "item.description"

    def to_coordinates(self, color):
        return {"text": self.text}

    @staticmethod
    def from_coordinates(coordinates):
        return coordinates["text"]

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            text = cls.from_coordinates(coordinates=_json["coordinates"])
        elif "data" in _json:
            text = cls.from_coordinates(coordinates=_json["data"])
        else:
            raise ValueError('Bad json, "coordinates", "data" or "description" not found')
        return cls(
            text=text,
        )



import logging
import re

from .. import entities, miscellaneous, exceptions, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name="dtlpy")


class Drivers:
    """
    Drivers Repository

    The Drivers class allows users to manage drivers that are used to connect with external storage.
    Read more about external storage in our `documentation <https://dataloop.ai/docs/overview-1>`_ and `developers' docs <https://developers.dataloop.ai/tutorials/data_management/>`_.
    """

    def __init__(self, client_api: ApiClient, project: entities.Project = None):
        """
        Initialize the Drivers repository.

        :param ApiClient client_api: The API client instance
        :param Project project: Optional project entity. If not provided, will try to get from checkout
        """
        self._client_api = client_api
        self._project = project

    ############
    # entities #
    ############
    @property
    def project(self) -> entities.Project:
        if self._project is None:
            # try get checkout
            project = self._client_api.state_io.get("project")
            if project is not None:
                self._project = entities.Project.from_json(_json=project, client_api=self._client_api)
        if self._project is None:
            raise exceptions.PlatformException(
                error="2001",
                message="Cannot perform action WITHOUT Project entity in Drivers repository."
                " Please checkout or set a project",
            )
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        """
        Set the project for this Drivers repository.

        :param Project project: The project entity to set
        :raises ValueError: If the provided project is not a valid Project entity
        """
        if not isinstance(project, entities.Project):
            raise ValueError("Must input a valid Project entity")
        self._project = project

    ###########
    # methods #
    ###########
    def __get_by_id(self, driver_id) -> entities.Driver:
        """
        Get a driver by its ID from the platform.

        :param str driver_id: The driver ID
        :return: Driver object
        :rtype: dtlpy.entities.driver.Driver
        :raises PlatformException: If the driver is not found or there's an API error
        """
        success, response = self._client_api.gen_request(req_type="get", path="/drivers/{}".format(driver_id))
        if success:
            _json = response.json()
            driver = self._getDriverClass(_json).from_json(client_api=self._client_api, _json=_json)
        else:
            raise exceptions.PlatformException(response)
        return driver

    @_api_reference.add(path="/drivers", method="get")
    def list(self) -> miscellaneous.List[entities.Driver]:
        """
        Get the project's drivers list.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :return: List of Drivers objects
        :rtype: list

        **Example**:

        .. code-block:: python

            project.drivers.list()

        """

        success, response = self._client_api.gen_request(
            req_type="get", path="/drivers?projectId={}".format(self.project.id)
        )
        if not success:
            raise exceptions.PlatformException(response)
        drivers = miscellaneous.List(
            [
                self._getDriverClass(_json).from_json(_json=_json, client_api=self._client_api)
                for _json in response.json()
            ]
        )
        return drivers

    def _getDriverClass(self, _json):
        """
        Determine the appropriate driver class based on the driver type in the JSON response.

        :param dict _json: The JSON response containing driver information
        :return: The appropriate driver class (S3Driver, GcsDriver, AzureBlobDriver, or Driver)
        :rtype: type
        """
        driver_type = _json.get("type", None)
        if driver_type == entities.ExternalStorage.S3:
            driver_class = entities.S3Driver
        elif driver_type == entities.ExternalStorage.GCS:
            driver_class = entities.GcsDriver
        elif driver_type in [entities.ExternalStorage.AZUREBLOB, entities.ExternalStorage.AZURE_DATALAKE_GEN2]:
            driver_class = entities.AzureBlobDriver
        else:
            driver_class = entities.Driver
        return driver_class

    @_api_reference.add(path="/drivers/{id}", method="get")
    def get(self, driver_name: str = None, driver_id: str = None) -> entities.Driver:
        """
        Get a Driver object to use in your code.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        You must provide at least ONE of the following params: driver_name, driver_id.

        :param str driver_name: optional - search by name
        :param str driver_id: optional - search by id
        :return: Driver object
        :rtype: dtlpy.entities.driver.Driver

        **Example**:

        .. code-block:: python

            project.drivers.get(driver_id='driver_id')
        """
        if driver_id is not None:
            driver = self.__get_by_id(driver_id)
        elif driver_name is not None:
            drivers = self.list()
            driver = [driver for driver in drivers if driver.name == driver_name]
            if not driver:
                # list is empty
                raise exceptions.PlatformException(
                    error="404", message="Driver not found. Name: {}".format(driver_name)
                )
                # driver = None
            elif len(driver) > 1:
                # more than one matching driver
                raise exceptions.PlatformException(
                    error="404", message='More than one driver with same name. Please "get" by id'
                )
            else:
                driver = driver[0]
        else:
            raise exceptions.PlatformException(error="400", message="Must provide an identifier (name or id) in inputs")
        return driver

    @_api_reference.add(path="/drivers", method="post")
    def _create_driver(self, payload: dict):
        """
        Create a driver on the platform using the provided payload.

        :param dict payload: The driver creation payload
        :return: The created driver object
        :rtype: dtlpy.entities.driver.Driver
        :raises PlatformException: If the driver creation fails
        """
        success, response = self._client_api.gen_request(req_type="post", path="/drivers", json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        else:
            _json = response.json()
            return self._getDriverClass(_json).from_json(_json=_json, client_api=self._client_api)

    def create_powerscale_s3(
        self,
        name: str,
        integration: entities.Integration,
        elastic_index: str,
        endpoint: str,
        bucket_name: str,
        region: str,
        elastic_index_path: str,
        path: str = None,
        project_id: str = None,
        allow_external_delete: bool = False
    ):
        """
        Create a PowerScale S3 driver.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str name: The driver name
        :param Integration integration: The S3 integration to use
        :param str elastic_index: The elastic index for PowerScale S3 driver
        :param str project_id: Optional project ID. If not provided, uses the current project
        :param bool allow_external_delete: True to allow deleting files from external storage when files are deleted in your Dataloop storage
        :param str endpoint: The endpoint URL. Must be in the format 'http://<hostname>:<port>' or 'https://<hostname>:<port>'
        :param str bucket_name: The external bucket name
        :param str region: The bucket region (relevant for S3)
        :param str elastic_index_path: The elastic index path for PowerScale S3 driver
        :param str path: The path to the NFS driver
        :return: The created PowerScale S3 driver object
        :rtype: dtlpy.entities.driver.Driver
        :raises ValueError: If the integration type is not S3
        :raises PlatformException: If the driver creation fails

        **Example**:

        .. code-block:: python

            project.drivers.create_powerscale_s3(
                name='powerscale_driver',
                integration=integration,
                elastic_index='my_index',
                elastic_index_path='my_index_path',
                endpoint='http://location:8765',
                bucket_name='my_bucket',
                region='us-west-1',
                path='my_path'
            )
        """
        if not integration.type == entities.IntegrationType.S3:
            raise ValueError("Integration type must be S3 for PowerScale S3 driver")

        payload = {
            "integrationId": integration.id,
            "integrationType": integration.type,
            "name": name,
            "metadata": {"system": {"projectId": self.project.id if project_id is None else project_id}},
            "type": entities.ExternalStorage.POWERSCALE_S3,
            "payload": {
                "elasticIndex": elastic_index, 
                "endpoint": endpoint,
                "path": path,
                "elasticIndexPath": elastic_index_path,
                "bucketName": bucket_name,
                "region": region
            },
            "allowExternalDelete": allow_external_delete,
            "creator": self._client_api.info().get("user_email"),
        }
        return self._create_driver(payload)


    def create_powerscale_nfs(
        self,
        name: str,
        elastic_index: str,
        elastic_index_path: str,
        integration: entities.Integration = None,
        project_id: str = None,
        allow_external_delete: bool = False
    ):
        """
        Create a PowerScale NFS driver.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str name: The driver name
        :param str elastic_index: The elastic index for PowerScale NFS driver
        :param Integration integration (optional): The S3 integration to use. If not provided, integration id will be 'system'
        :param str project_id: Optional project ID. If not provided, uses the current project
        :param bool allow_external_delete: True to allow deleting files from external storage when files are deleted in your Dataloop storage
        :param str elastic_index_path: The elastic index path for PowerScale NFS driver
        :return: The created PowerScale NFS driver object
        :rtype: dtlpy.entities.driver.Driver
        :raises PlatformException: If the driver creation fails

        **Example**:

        .. code-block:: python

            project.drivers.create_powerscale_nfs(
                name='powerscale_driver',
                elastic_index='my_index',
                elastic_index_path='my_index_path',
            )
        """

        integration_id = 'system'
        integration_type = None
        if integration is not None:
            if not integration.type == entities.IntegrationType.S3:
                raise ValueError("Integration type must be S3 for PowerScale NFS driver")
            integration_id = integration.id
            integration_type = integration.type

        payload = {
            "integrationId": integration_id,
            "integrationType": integration_type,
            "name": name,
            "metadata": {"system": {"projectId": self.project.id if project_id is None else project_id}},
            "type": entities.ExternalStorage.POWERSCALE_NFS,
            "payload": {"elasticIndex": elastic_index, "elasticIndexPath": elastic_index_path},
            "allowExternalDelete": allow_external_delete,
            "creator": self._client_api.info().get("user_email"),
        }
        return self._create_driver(payload)

    def create_min_io(
        self,
        name: str,
        integration: entities.Integration,
        bucket_name: str,
        endpoint: str,
        project_id: str = None,
        allow_external_delete: bool = True,
        region: str = None,
        storage_class: str = "",
        path: str = "",
    ):
        """
        Create a MinIO driver.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str name: The driver name
        :param Integration integration: The S3 integration to use
        :param str bucket_name: The external bucket name
        :param str endpoint: The MinIO endpoint URL. Must be in the format 'http://<hostname>:<port>' or 'https://<hostname>:<port>'
        :param str project_id: Optional project ID. If not provided, uses the current project
        :param bool allow_external_delete: True to allow deleting files from external storage when files are deleted in your Dataloop storage
        :param str region: The bucket region (relevant for S3)
        :param str storage_class: The storage class (relevant for S3)
        :param str path: Optional path. By default path is the root folder. Path is case sensitive
        :return: The created MinIO driver object
        :rtype: dtlpy.entities.driver.Driver
        :raises ValueError: If the integration type is not S3 or if the endpoint URL format is invalid
        :raises PlatformException: If the driver creation fails

        **Example**:

        .. code-block:: python

            project.drivers.create_min_io(
                name='minio_driver',
                integration=integration,
                bucket_name='my_bucket',
                endpoint='http://localhost:9000'
            )
        """
        if not integration.type == entities.IntegrationType.S3:
            raise ValueError("Integration type must be S3 for Minio driver")
        
        if not re.match(r"^https?://[A-Za-z0-9.-]+:\d+$", endpoint):
            raise ValueError(
                f"Invalid endpoint URL '{endpoint}'. Must be 'http://<hostname>:<port>' or 'https://<hostname>:<port>'."
            )
        
        payload = {
            "integrationId": integration.id,
            "integrationType": integration.type,
            "name": name,
            "metadata": {"system": {"projectId": self.project.id if project_id is None else project_id}},
            "type": entities.ExternalStorage.MIN_IO,
            "payload": {"bucketName": bucket_name, "storageClass": storage_class, "region": region, "path": path, "endpoint": endpoint},
            "allowExternalDelete": allow_external_delete,
            "creator": self._client_api.info().get("user_email"),
        }

        return self._create_driver(payload)


    def create(
        self,
        name: str,
        driver_type: entities.ExternalStorage,
        integration_id: str,
        bucket_name: str,
        integration_type: entities.IntegrationType,
        project_id: str = None,
        allow_external_delete: bool = True,
        region: str = None,
        storage_class: str = "",
        path: str = "",
        endpoint: str = None
    ):
        """
        Create a storage driver.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str name: the driver name
        :param ExternalStorage driver_type: dl.ExternalStorage (Enum). For all options run: list(dl.ExsternalStorage)
        :param str integration_id: the integration id
        :param str bucket_name: the external bucket name
        :param IntegrationType integration_type: dl.IntegrationType (Enum). For all options run: list(dl.IntegrationType)
        :param str project_id: project id
        :param bool allow_external_delete: true to allow deleting files from external storage when files are deleted in your Dataloop storage
        :param str region: relevant only for s3 - the bucket region
        :param str storage_class: relevant only for s3
        :param str path: Optional. By default path is the root folder. Path is case sensitive integration
        :param str endpoint: Optional. Custom endpoint for minio storage. Must be in the format 'http://<hostname>:<port>' or 'https://<hostname>:<port>'.
        :return: driver object
        :rtype: dtlpy.entities.driver.Driver

        **Example**:

        .. code-block:: python

            project.drivers.create(name='driver_name',
                       driver_type=dl.ExternalStorage.S3,
                       integration_id='integration_id',
                       bucket_name='bucket_name',
                       project_id='project_id',
                       region='ey-west-1')
        """
        if integration_type is None:
            integration_type = driver_type
        if driver_type == entities.ExternalStorage.S3:
            bucket_payload = "bucketName"
            if endpoint:
                if not re.match(r"^https?://[A-Za-z0-9.-]+:\d+$", endpoint):
                    raise ValueError(
                        f"Invalid endpoint URL '{endpoint}'. Must be 'http://<hostname>:<port>' or 'https://<hostname>:<port>'."
                    )

        elif driver_type == entities.ExternalStorage.GCS:
            bucket_payload = "bucket"
        else:
            bucket_payload = "containerName"
        payload = {
            "integrationId": integration_id,
            "integrationType": integration_type,
            "name": name,
            "metadata": {"system": {"projectId": self.project.id if project_id is None else project_id}},
            "type": driver_type,
            "payload": {bucket_payload: bucket_name, "storageClass": storage_class, "region": region, "path": path},
            "allowExternalDelete": allow_external_delete,
            "creator": self._client_api.info().get("user_email"),
        }

        if endpoint and driver_type == entities.ExternalStorage.S3:
            payload["payload"]["endpoint"] = endpoint

        return self._create_driver(payload)
    
    @_api_reference.add(path="/drivers/{id}", method="delete")
    def delete(self, driver_name: str = None, driver_id: str = None, sure: bool = False, really: bool = False):
        """
        Delete a driver forever!

        **Prerequisites**: You must be an *owner* or *developer* to use this method.

        **Example**:

        .. code-block:: python

            project.drivers.delete(dataset_id='dataset_id', sure=True, really=True)

        :param str driver_name: optional - search by name
        :param str driver_id: optional - search by id
        :param bool sure: Are you sure you want to delete?
        :param bool really: Really really sure?
        :return: True if success
        :rtype: bool
        """
        if sure and really:
            driver = self.get(driver_name=driver_name, driver_id=driver_id)
            success, response = self._client_api.gen_request(req_type="delete", path="/drivers/{}".format(driver.id))
            if not success:
                raise exceptions.PlatformException(response)
            logger.info("Driver {!r} was deleted successfully".format(driver.name))
            return True
        else:
            raise exceptions.PlatformException(
                error="403", message="Cant delete driver from SDK. Please login to platform to delete"
            )



def main():
    """
    Convert yolo annotation and images to Dataloop
    Yolo annotations format:
    For each image there is a text file with same name that has a list of box location and label index
    """
    import dtlpy as dl
    from PIL import Image
    import os

    # Get project and dataset
    project = dl.projects.get(project_name='Fruits')
    dataset = project.datasets.get(dataset_name='Rambutan')

    images_and_annotations_path = '/home/fruits/data'

    # read all images from local dataset
    img_filepaths = list()
    for path, subdirs, files in os.walk(images_and_annotations_path):
        for filename in files:
            striped, ext = os.path.splitext(filename)
            if ext in ['.jpeg']:
                img_filepaths.append(os.path.join(path, filename))

    classes_filepath = '/home/fruits/classes.txt'
    # get labels from file
    with open(classes_filepath, 'r') as f:
        labels = {i_line: label.strip() for i_line, label in enumerate(f.readlines())}

    for filepath in img_filepaths:
        # get image height and width
        img = Image.open(filepath)
        width, height = img.size()
        # upload item to platform
        item = dataset.items.update(filepath=filepath)

        # get YOLO annotations
        _, ext = os.path.splitext(filepath)
        annotation_filepath = filepath.replace(ext, '.txt')
        with open(annotation_filepath, 'r') as f:
            annotations = f.read().split('\n')

        builder = item.annotations.builder()
        # convert to Dataloop annotations
        for annotation in annotations:
            if not annotation:
                continue
            # convert annotation format
            elements = annotation.split(" ")
            label_id = elements[0]

            xmin_add_xmax = float(elements[1]) * (2.0 * float(width))
            ymin_add_ymax = float(elements[2]) * (2.0 * float(height))

            w = float(elements[3]) * float(width)
            h = float(elements[4]) * float(height)

            left = (xmin_add_xmax - w) / 2
            top = (ymin_add_ymax - h) / 2
            right = left + w
            bottom = top + h

            # add to annotations
            builder.add(annotation_definition=dl.Box(top=top,
                                                     left=left,
                                                     bottom=bottom,
                                                     right=right,
                                                     label=labels[label_id]))
            # upload all annotations of an item
            builder.upload()



import logging
import traceback

from .. import entities, miscellaneous, exceptions, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Ontologies:
    """
    Ontologies Repository

    The Ontologies class allows users to manage ontologies and their properties.
    Read more about ontology in our `developers docs <https://developers.dataloop.ai/tutorials/recipe_and_ontology/ontology/chapter/>`_.
    """

    def __init__(self, client_api: ApiClient,
                 recipe: entities.Recipe = None,
                 project: entities.Project = None,
                 dataset: entities.Dataset = None):
        self._client_api = client_api
        self._recipe = recipe
        self._project = project
        self._dataset = dataset

    ############
    # entities #
    ############
    @property
    def recipe(self) -> entities.Recipe:
        if self._recipe is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "recipe". need to set a Recipe entity or use ontology.recipes repository')
        assert isinstance(self._recipe, entities.Recipe)
        return self._recipe

    @recipe.setter
    def recipe(self, recipe: entities.Recipe):
        if not isinstance(recipe, entities.Recipe):
            raise ValueError('Must input a valid Recipe entity')
        self._recipe = recipe

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "project". need to set a Project entity or use project.ontologies repository')
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "dataset". need to set a Dataset entity or use dataset.ontologies repository')
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        if not isinstance(dataset, entities.Dataset):
            raise ValueError('Must input a valid Dataset entity')
        self._dataset = dataset

    def __get_project_ids(self, project_ids):
        if project_ids is not None:
            return project_ids if isinstance(project_ids, list) else [project_ids]
        elif self._recipe is not None:
            return self._recipe.project_ids
        elif self._project is not None:
            return [self._project.id]
        elif self._dataset is not None:
            return self._dataset.projects
        else:
            return project_ids

    ###########
    # methods #
    ###########
    @_api_reference.add(path='/ontologies', method='post')
    def create(self,
               labels,
               title=None,
               project_ids=None,
               attributes=None
               ) -> entities.Ontology:
        """
        Create a new ontology.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param labels: recipe tags
        :param str title: ontology title, name
        :param list project_ids: recipe project/s
        :param list attributes: recipe attributes
        :return: Ontology object
        :rtype: dtlpy.entities.ontology.Ontology

        **Example**:

        .. code-block:: python

            recipe.ontologies.create(labels='labels_entity',
                                  title='new_ontology',
                                  project_ids='project_ids')
        """
        project_ids = self.__get_project_ids(project_ids=project_ids)
        if attributes is None:
            attributes = list()
        elif not isinstance(project_ids, list):
            project_ids = [project_ids]
        # convert to platform label format (root)
        labels = self.labels_to_roots(labels)
        payload = {"roots": labels,
                   "projectIds": project_ids,
                   "attributes": attributes}
        if title is not None:
            payload['title'] = title
        success, response = self._client_api.gen_request(req_type="post",
                                                         path="/ontologies",
                                                         json_req=payload)
        if success:
            logger.info("Ontology was created successfully")
            ontology = entities.Ontology.from_json(_json=response.json(),
                                                   client_api=self._client_api,
                                                   recipe=self._recipe)
            if self._recipe:
                self._recipe.ontology_ids.append(ontology.id)
                self._recipe.update()
        else:
            logger.error("Failed to create Ontology")
            raise exceptions.PlatformException(response)
        return ontology

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Ontology]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_package, package in enumerate(response_items):
            jobs[i_package] = pool.submit(entities.Ontology._protected_from_json,
                                          **{'client_api': self._client_api,
                                             '_json': package,
                                             'project': self._project,
                                             'dataset': self._dataset,
                                             'recipe': self._recipe})

        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        ontologies = miscellaneous.List([r[1] for r in results if r[0] is True])
        return ontologies

    def _list(self, filters: entities.Filters):
        url = '/ontologies?pageOffset={}&pageSize={}'.format(filters.page, filters.page_size)
        project_ids = None
        ids = None
        for single_filter in filters.and_filter_list:
            if single_filter.field == 'projects':
                project_ids = single_filter.values
                break
        for single_filter in filters.and_filter_list:
            if single_filter.field == 'ids':
                ids = single_filter.values
                break

        if project_ids:
            url = '{}&projects={}'.format(url, ','.join(project_ids))
        if ids:
            url = '{}&ids={}'.format(url, ','.join(ids))

        # request
        success, response = self._client_api.gen_request(req_type='get',
                                                         path=url)
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def __list(self, filters: entities.Filters) -> entities.PagedEntities:
        """
        List project ontologies.
        
        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return:
        """
        paged = entities.PagedEntities(items_repository=self,
                                       filters=filters,
                                       page_offset=filters.page,
                                       page_size=filters.page_size,
                                       client_api=self._client_api)
        paged.get_page()
        return paged

    def list(self, project_ids=None) -> miscellaneous.List[entities.Ontology]:
        """
        List ontologies for recipe

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param project_ids:
        :return: list of all the ontologies

        **Example**:

        .. code-block:: python

            recipe.ontologies.list(project_ids='project_ids')
        """
        if self._recipe is not None:
            ontologies = [ontology_id for ontology_id in self.recipe.ontology_ids]

            pool = self._client_api.thread_pools(pool_name='entity.create')
            jobs = [None for _ in range(len(ontologies))]
            for i_ontology, ontology_id in enumerate(ontologies):
                jobs[i_ontology] = pool.submit(self._protected_get, **{'ontology_id': ontology_id})

            # get all results
            results = [j.result() for j in jobs]
            # log errors
            _ = [logger.warning(r[1]) for r in results if r[0] is False]
            # return good jobs
            return miscellaneous.List([r[1] for r in results if r[0] is True])
        else:
            filters = entities.Filters(resource=entities.FiltersResource.ONTOLOGY)
            project_ids = self.__get_project_ids(project_ids=project_ids)
            if project_ids:
                filters.add(field='projects', values=self.__get_project_ids(project_ids=project_ids))
            if self._dataset:
                filters.add(field='ids', values=self._dataset.ontology_ids)
            ontologies = list()
            pages = self.__list(filters=filters)
            for page in pages:
                ontologies += page
            return miscellaneous.List(ontologies)

    def _protected_get(self, ontology_id):
        """
        Same as get but with try-except to catch if error
        :param ontology_id:
        :return:
        """
        try:
            ontology = self.get(ontology_id=ontology_id)
            status = True
        except Exception:
            ontology = traceback.format_exc()
            status = False
        return status, ontology

    @_api_reference.add(path='/ontologies/{id}', method='get')
    def get(self, ontology_id: str) -> entities.Ontology:
        """
        Get Ontology object to use in your code.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str ontology_id: ontology id
        :return: Ontology object
        :rtype: dtlpy.entities.ontology.Ontology

        **Example**:

        .. code-block:: python

            recipe.ontologies.get(ontology_id='ontology_id')
        """
        success, response = self._client_api.gen_request(req_type="get",
                                                         path="/ontologies/{}".format(ontology_id))
        if success:
            ontology = entities.Ontology.from_json(_json=response.json(),
                                                   client_api=self._client_api,
                                                   recipe=self._recipe,
                                                   dataset=self._dataset,
                                                   project=self._project)
        else:
            raise exceptions.PlatformException(response)
        return ontology

    @_api_reference.add(path='/ontologies/{id}', method='delete')
    def delete(self, ontology_id):
        """
        Delete Ontology from the platform.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param ontology_id: ontology id
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            recipe.ontologies.delete(ontology_id='ontology_id')
        """
        success, response = self._client_api.gen_request(req_type="delete",
                                                         path="/ontologies/%s" % ontology_id)
        if success:
            logger.debug("Ontology was deleted successfully")
            return success
        else:
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/ontologies/{id}', method='put')
    def update(self, ontology: entities.Ontology, system_metadata=False) -> entities.Ontology:
        """
        Update the Ontology metadata.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

       :param dtlpy.entities.ontology.Ontology ontology: Ontology object
       :param bool system_metadata: bool - True, if you want to change metadata system
       :return: Ontology object
       :rtype: dtlpy.entities.ontology.Ontology

       **Example**:

        .. code-block:: python

            recipe.ontologies.delete(ontology='ontology_entity')
       """
        url_path = "/ontologies/%s" % ontology.id
        if system_metadata:
            url_path += "?system=true"
        success, response = self._client_api.gen_request(req_type="put",
                                                         path=url_path,
                                                         json_req=ontology.to_json())
        if success:
            logger.debug("Ontology was updated successfully")
            # update dataset labels
            ontology = entities.Ontology.from_json(_json=response.json(),
                                                   client_api=self._client_api,
                                                   recipe=self._recipe)
            if self._recipe is not None and self._recipe._dataset is not None:
                self.recipe.dataset._labels = ontology.labels
            return ontology
        else:
            logger.error("Failed to update ontology:")
            raise exceptions.PlatformException(response)

    @staticmethod
    def labels_to_roots(labels):
        """
        Converts labels dictionary to a list of platform representation of labels.

        :param dict labels: labels dict
        :return: platform representation of labels
        """
        roots = list()
        if isinstance(labels, dict):
            for label in labels:
                root = {
                    "value": {
                        "tag": label,
                        "color": labels[label],
                        "attributes": list(),
                    },
                    "children": list(),
                }
                roots.append(root)
        elif isinstance(labels, list):
            for label in labels:
                if isinstance(label, entities.Label):
                    root = label.to_root()
                elif "value" in label:
                    root = {
                        "value": label["value"],
                        "children": label.get("children", list()),
                    }
                else:
                    root = {
                        "value": {
                            "tag": label.get("tag", None),
                            "color": label.get("color", "#FFFFFF"),
                            "attributes": label.get("attributes", list()),
                        },
                        "children": label.get("children", list()),
                    }
                roots.append(root)
        for root in roots:
            if not isinstance(root["value"]["color"], str):
                # noinspection PyStringFormat
                root["value"]["color"] = "#%02x%02x%02x" % root["value"]["color"]
        return roots

    def update_attributes(self,
                          ontology_id: str,
                          title: str,
                          key: str,
                          attribute_type: entities.AttributesTypes,
                          scope: list = None,
                          optional: bool = None,
                          values: list = None,
                          attribute_range: entities.AttributesRange = None):
        """
        ADD a new attribute or update if exist

        :param str ontology_id: ontology_id
        :param str title: attribute title
        :param str key: the key of the attribute must br unique
        :param AttributesTypes attribute_type: dl.AttributesTypes your attribute type
        :param list scope: list of the labels or * for all labels
        :param bool optional: optional attribute
        :param list values: list of the attribute values ( for checkbox and radio button)
        :param dict or AttributesRange attribute_range: dl.AttributesRange object
        :return: true in success
        :rtype: bool

        **Example**:

        .. code-block:: python

            ontology.update_attributes(key='1',
                                       title='checkbox',
                                       attribute_type=dl.AttributesTypes.CHECKBOX,
                                       values=[1,2,3])
        """
        if not title:
            raise exceptions.PlatformException(400, "title must be provided")
        url_path = '/ontologies/{ontology_id}/attributes'.format(ontology_id=ontology_id)

        multi = None
        if attribute_type == entities.AttributesTypes.CHECKBOX:
            attribute_type = 'options'
            multi = True
        elif attribute_type == entities.AttributesTypes.RADIO_BUTTON:
            attribute_type = 'options'
            multi = False

        # build attribute json
        attribute_json = {
            'title': title,
            'key': key,
            'type': attribute_type,
        }

        if optional is not None:
            attribute_json['optional'] = optional

        if multi is not None:
            attribute_json['multi'] = multi

        if values is not None:
            if not isinstance(values, list):
                values = [values]
            for val in values:
                if not isinstance(val, str):
                    raise exceptions.PlatformException(400, 'Attributes values type must be list of strings')
            attribute_json['values'] = values

        if attribute_range is not None:
            attribute_json['range'] = attribute_range.to_json()

        if scope is not None:
            if not isinstance(scope, list):
                scope = [scope]
        else:
            scope = ['*']
        attribute_json['scope'] = scope

        json_req = {
            'items': [attribute_json],
            'upsert': True
        }

        success, response = self._client_api.gen_request(req_type="PATCH",
                                                         path=url_path,
                                                         json_req=json_req)
        if not success:
            raise exceptions.PlatformException(response)
        return True

    def delete_attributes(self, ontology_id, keys: list):
        """
        Delete a bulk of attributes

        :param str ontology_id: ontology id
        :param list keys: Keys of attributes to delete
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            ontology.delete_attributes(['1'])
        """

        if not isinstance(keys, list):
            keys = [keys]
        url_path = '/ontologies/{ontology_id}/attributes'.format(ontology_id=ontology_id)
        json_req = {
            'keys': keys
        }
        success, response = self._client_api.gen_request(req_type="DELETE",
                                                         path=url_path,
                                                         json_req=json_req)
        if not success:
            raise exceptions.PlatformException(response)
        return True
